from .id import Id
from .util import log
from .live2d_gl_wrapper import Live2DGLWrapper

class Live2D:
    L2D_OUTSIDE_PARAM_AVAILABLE = False
    __firstInit = True
    clippingMaskBufferSize = 256

    @staticmethod
    def init():
        if Live2D.__firstInit:
            Live2D.__firstInit = False

    @staticmethod
    def clearBuffer():
        Live2DGLWrapper.clearColor(0, 0, 0, 0)
        Live2DGLWrapper.clear(Live2DGLWrapper.COLOR_BUFFER_BIT)

    @staticmethod
    def dispose():
        Id.releaseStored()



class StandardParams:
    ParamAngleX = "ParamAngleX"
    ParamAngleY = "ParamAngleY"
    ParamAngleZ = "ParamAngleZ"
    ParamEyeLOpen = "ParamEyeLOpen"
    ParamEyeLSmile = "ParamEyeLSmile"
    ParamEyeROpen = "ParamEyeROpen"
    ParamEyeRSmile = "ParamEyeRSmile"
    ParamEyeBallX = "ParamEyeBallX"
    ParamEyeBallY = "ParamEyeBallY"
    ParamEyeBallForm = "ParamEyeBallForm"
    ParamBrowLY = "ParamBrowLY"
    ParamBrowRY = "ParamBrowRY"
    ParamBrowLX = "ParamBrowLX"
    ParamBrowRX = "ParamBrowRX"
    ParamBrowLAngle = "ParamBrowLAngle"
    ParamBrowRAngle = "ParamBrowRAngle"
    ParamBrowLForm = "ParamBrowLForm"
    ParamBrowRForm = "ParamBrowRForm"
    ParamMouthForm = "ParamMouthForm"
    ParamMouthOpenY = "ParamMouthOpenY"
    ParamCheek = "ParamCheek"
    ParamBodyAngleX = "ParamBodyAngleX"
    ParamBodyAngleY = "ParamBodyAngleY"
    ParamBodyAngleZ = "ParamBodyAngleZ"
    ParamBreath = "ParamBreath"
    ParamArmLA = "ParamArmLA"
    ParamArmRA = "ParamArmRA"
    ParamArmLB = "ParamArmLB"
    ParamArmRB = "ParamArmRB"
    ParamHandL = "ParamHandL"
    ParamHandR = "ParamHandR"
    ParamHairFront = "ParamHairFront"
    ParamHairSide = "ParamHairSide"
    ParamHairBack = "ParamHairBack"
    ParamHairFluffy = "ParamHairFluffy"
    ParamShoulderY = "ParamShoulderY"
    ParamBustX = "ParamBustX"
    ParamBustY = "ParamBustY"
    ParamBaseX = "ParamBaseX"
    ParamBaseY = "ParamBaseY"


class Parameter:

    def __init__(self):
        self.id: str = ""
        self.type: int = 0
        self.value: float = 0
        self.max: float = 0
        self.min: float = 0
        self.default: float = 0
    
    def __str__(self):
        return f"Parameter(id={self.id}, type={self.type}, value={self.value}, max={self.max}, min={self.min}, default={self.default})"


from ..type import Array, Float32Array
from .clip_rectf import ClipRectF
from .clip_draw_context import ClipDrawContext


class ClipContext:

    def __init__(self, aH, aK, aI):
        self.clipIDList = Array()
        self.clipIDList = aI
        self.clippingMaskDrawIndexList = Array()
        for aJ in range(0, len(aI), 1):
            self.clippingMaskDrawIndexList.append(aK.getDrawDataIndex(aI[aJ]))

        self.clippedDrawContextList = Array()
        self.isUsing = True
        self.layoutChannelNo = 0
        self.layoutBounds = ClipRectF()
        self.allClippedDrawRect = ClipRectF()
        self.matrixForMask = Float32Array(16)
        self.matrixForDraw = Float32Array(16)
        self.owner = aH

    def addClippedDrawData(self, aJ, aI):
        aH = ClipDrawContext(aJ, aI)
        self.clippedDrawContextList.append(aH)



import time


class UtSystem:
    USER_TIME_AUTO = -1
    userTimeMSec = USER_TIME_AUTO

    @staticmethod
    def isBigEndian():
        return True

    @staticmethod
    def wait(duration):
        try:
            start_time = UtSystem.getTimeMSec()
            while UtSystem.getTimeMSec() - start_time < duration:
                pass

        except Exception as e:
            print(e)

    @staticmethod
    def getUserTimeMSec():
        return UtSystem.getSystemTimeMSec() if (
                    UtSystem.userTimeMSec == UtSystem.USER_TIME_AUTO) else UtSystem.userTimeMSec

    @staticmethod
    def setUserTimeMSec(aH):
        UtSystem.userTimeMSec = aH

    @staticmethod
    def updateUserTimeMSec():
        UtSystem.userTimeMSec = UtSystem.getSystemTimeMSec()
        return UtSystem.userTimeMSec

    @staticmethod
    def getTimeMSec():
        return time.time() * 1000

    @staticmethod
    def getSystemTimeMSec():
        return time.time() * 1000

    @staticmethod
    def arraycopy(aM, aJ, aI, aL, aH):
        for aK in range(0, aH, 1):
            aI[aL + aK] = aM[aJ + aK]



import math

from ...core import UtSystem


class L2DTargetPoint:
    FRAME_RATE = 30
    TIME_TO_MAX_SPEED = 0.15
    FACE_PARAM_MAX_V = 40.0 / 7.5
    MAX_V = FACE_PARAM_MAX_V / FRAME_RATE
    FRAME_TO_MAX_SPEED = TIME_TO_MAX_SPEED * FRAME_RATE

    def __init__(self):
        self.EPSILON = 0.01
        self.faceTargetX = 0
        self.faceTargetY = 0
        self.faceX = 0
        self.faceY = 0
        self.faceVX = 0
        self.faceVY = 0
        self.lastTimeSec = 0

    def setPoint(self, x, y):
        self.faceTargetX = x
        self.faceTargetY = y

    def getX(self):
        return self.faceX

    def getY(self):
        return self.faceY

    def update(self):

        if self.lastTimeSec == 0:
            self.lastTimeSec = UtSystem.getUserTimeMSec()
            return

        cur_time_sec = UtSystem.getUserTimeMSec()
        delta_time_weight = (cur_time_sec - self.lastTimeSec) * L2DTargetPoint.FRAME_RATE / 1000.0
        self.lastTimeSec = cur_time_sec
        max_a = delta_time_weight * self.MAX_V / self.FRAME_TO_MAX_SPEED
        dx = (self.faceTargetX - self.faceX)
        dy = (self.faceTargetY - self.faceY)
        if abs(dx) <= self.EPSILON and abs(dy) <= self.EPSILON:
            return
        d = math.sqrt(dx * dx + dy * dy)
        vx = self.MAX_V * dx / d
        vy = self.MAX_V * dy / d
        ax = vx - self.faceVX
        ay = vy - self.faceVY
        a = math.sqrt(ax * ax + ay * ay)
        if a < -max_a or a > max_a:
            ax *= max_a / a
            ay *= max_a / a

        self.faceVX += ax
        self.faceVY += ay

        max_v = 0.5 * (math.sqrt(max_a * max_a + 16 * max_a * d - 8 * max_a * d) - max_a)
        cur_v = math.sqrt(self.faceVX * self.faceVX + self.faceVY * self.faceVY)
        if cur_v > max_v:
            self.faceVX *= max_v / cur_v
            self.faceVY *= max_v / cur_v

        self.faceX += self.faceVX
        self.faceY += self.faceVY



from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from ..model_context import ModelContext

from ..DEF import GOSA, PIVOT_TABLE_SIZE
from ..io.iserializable import ISerializable
from ..param import ParamPivots


class PivotManager(ISerializable):

    def __init__(self):
        self.paramPivotTable = None

    def read(self, aH):
        self.paramPivotTable = aH.readObject()

    def checkParamUpdated(self, aK):
        if aK.requireSetup():
            return True

        aH = aK.getInitVersion()
        for aJ in range(len(self.paramPivotTable) - 1, -1, -1):
            aI = self.paramPivotTable[aJ].getParamIndex(aH)
            if aI == ParamPivots.PARAM_INDEX_NOT_INIT:
                aI = aK.getParamIndex(self.paramPivotTable[aJ].getParamID())

            if aK.isParamUpdated(aI):
                return True

        return False

    def calcPivotValues(self, mdc: 'ModelContext', ret: List[bool]):
        aX = len(self.paramPivotTable)
        aJ = mdc.getInitVersion()
        aN = 0
        for aK in range(0, aX, 1):
            aH = self.paramPivotTable[aK]
            aI = aH.getParamIndex(aJ)
            if aI == ParamPivots.PARAM_INDEX_NOT_INIT:
                aI = mdc.getParamIndex(aH.getParamID())
                aH.setParamIndex(aI, aJ)

            if aI < 0:
                raise Exception("err 23242 : " + aH.getParamID())

            aU = 0 if aI < 0 else mdc.getParamFloat(aI)
            aQ = aH.getPivotCount()
            aM = aH.getPivotValues()
            aP = -1
            aT = 0
            if aQ < 1:
                pass
            else:
                if aQ == 1:
                    aS = aM[0]
                    if aS - GOSA < aU < aS + GOSA:
                        aP = 0
                        aT = 0
                    else:
                        aP = 0
                        ret[0] = True
                else:
                    aS = aM[0]
                    if aU < aS - GOSA:
                        aP = 0
                        # TODO: 不清楚原因，需要找其他模型测试
                        # 其他分支对ret的赋值有可能导致问题
                        if isinstance(ret, list): # by heurry@Github
                            ret[0] = True
                        else:
                            ret = [True]
                    else:
                        if aU < aS + GOSA:
                            aP = 0
                        else:
                            aW = False
                            for aO in range(1, aQ, 1):
                                aR = aM[aO]
                                if aU < aR + GOSA:
                                    if aR - GOSA < aU:
                                        aP = aO
                                    else:
                                        aP = aO - 1
                                        aT = (aU - aS) / (aR - aS)
                                        aN += 1

                                    aW = True
                                    break

                                aS = aR

                            if not aW:
                                aP = aQ - 1
                                aT = 0
                                ret[0] = True

            aH.setTmpPivotIndex(aP)
            aH.setTmpT(aT)

        return aN

    def calcPivotIndices(self, aN, aT, aP):
        aR = 1 << aP
        if aR + 1 > PIVOT_TABLE_SIZE:
            print("err 23245\n")

        aS = len(self.paramPivotTable)
        aK = 1
        aH = 1
        aJ = 0
        for aQ in range(0, aR, 1):
            aN[aQ] = 0

        for aL in range(0, aS, 1):
            aI = self.paramPivotTable[aL]
            if aI.getTmpT() == 0:
                aO = aI.getTmpPivotIndex() * aK
                if aO < 0:
                    raise Exception("err 23246")

                for aQ in range(0, aR, 1):
                    aN[aQ] += aO
            else:
                aO = aK * aI.getTmpPivotIndex()
                aM = aK * (aI.getTmpPivotIndex() + 1)
                for aQ in range(0, aR, 1):
                    aN[aQ] += aO if (int(aQ / aH) % 2 == 0) else aM

                aT[aJ] = aI.getTmpT()
                aJ += 1
                aH *= 2

            aK *= aI.getPivotCount()

        aN[aR] = 65535
        aT[aJ] = -1

    def getParamCount(self):
        return len(self.paramPivotTable)



from typing import List, Optional, Any


def Float32Array(size: int) -> List[float]:
    if not isinstance(size, int):
        raise Exception("invalid param")

    return [0.0] * size


def Float64Array(size: int) -> List[float]:
    return Float32Array(size)


def Int8Array(size: int) -> List[int]:
    return Int32Array(size)


def Int16Array(size: int) -> List[int]:
    return Int32Array(size)


def Int32Array(size: Optional[int] = None) -> List[int]:
    if size is None:
        return []
    if not isinstance(size, int):
        raise Exception("invalid param")

    return [0] * size


def Array(size: Optional[int] = None) -> List[Any]:
    if size is None:
        return []
    if not isinstance(size, int):
        raise Exception("invalid param")

    return [None] * size



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Union

from .DEF import LIVE2D_FORMAT_VERSION_AVAILABLE, LIVE2D_FORMAT_VERSION_V2_8_TEX_OPTION
from .id import Id
from .io import BinaryReader
from .model import ModelImpl
from .model_context import ModelContext

if TYPE_CHECKING:
    from .draw import MeshContext, IDrawData, Mesh


class ALive2DModel(ABC):

    def __init__(self):
        self.modelImpl = None
        self.modelContext = None
        self.modelContext = ModelContext(self)

    def setModelImpl(self, moc: 'ModelImpl'):
        self.modelImpl = moc

    def getModelImpl(self):
        if self.modelImpl is None:
            self.modelImpl = ModelImpl()
            self.modelImpl.initDirect()

        return self.modelImpl

    def getCanvasWidth(self) -> int:
        if self.modelImpl is None:
            return 0

        return self.modelImpl.getCanvasWidth()

    def getCanvasHeight(self):
        if self.modelImpl is None:
            return 0

        return self.modelImpl.getCanvasHeight()

    def getParamFloat(self, x):
        if not isinstance(x, int):
            x = self.modelContext.getParamIndex(Id.getID(x))

        return self.modelContext.getParamFloat(x)

    def setParamFloat(self, x: Union[int, str], value: float, weight: float = 1):
        if not isinstance(x, int):
            x = self.modelContext.getParamIndex(Id.getID(x))
        value = 0 if value is None else value
        self.modelContext.setParamFloat(x, self.modelContext.getParamFloat(x) * (1 - weight) + value * weight)

    def addToParamFloat(self, x: Union[int, str], value: float, weight: float = 1):
        if not isinstance(x, int):
            x = self.modelContext.getParamIndex(Id.getID(x))

        self.modelContext.setParamFloat(x, self.modelContext.getParamFloat(x) + value * weight)

    def multParamFloat(self, x: Union[int, str], value: float, weight: float = 1):
        if not isinstance(x, int):
            x = self.modelContext.getParamIndex(Id.getID(x))

        self.modelContext.setParamFloat(x, self.modelContext.getParamFloat(x) * (1 + (value - 1) * weight))

    def getParamIndex(self, idStr: str) -> int:
        return self.modelContext.getParamIndex(Id.getID(idStr))

    def loadParam(self):
        self.modelContext.loadParam()

    def saveParam(self):
        self.modelContext.saveParam()

    def init(self):
        self.modelContext.init()

    def update(self):
        self.modelContext.update()

    @abstractmethod
    def draw(self):
        pass

    def getModelContext(self):
        return self.modelContext

    def setPartsOpacity(self, index: Union[int, str], opacity: float):
        if not isinstance(index, int):
            index = self.modelContext.getPartsDataIndex(Id.getID(index))

        self.modelContext.setPartsOpacity(index, opacity)

    def getPartsDataIndex(self, aH):
        if not (isinstance(aH, Id)):
            aH = Id.getID(aH)

        return self.modelContext.getPartsDataIndex(aH)

    def getPartsOpacity(self, partIndex):
        if not isinstance(partIndex, int):
            partIndex = self.modelContext.getPartsDataIndex(Id.getID(partIndex))

        if partIndex < 0:
            return 0

        return self.modelContext.getPartsOpacity(partIndex)

    @abstractmethod
    def getDrawParam(self):
        pass

    def getDrawDataIndex(self, drawId: str):
        if not isinstance(drawId, str):
            raise RuntimeError
        return self.modelContext.getDrawDataIndex(Id.getID(drawId))

    def getDrawData(self, aH):
        return self.modelContext.getDrawData(aH)

    def getTransformedPoints(self, aH):
        aI = self.modelContext.getDrawContext(aH)
        if isinstance(aI, MeshContext):
            return aI.getTransformedPoints()

        return None

    def getIndexArray(self, aI):
        if aI < 0 or aI >= len(self.modelContext.drawDataList):
            return None

        aH = self.modelContext.drawDataList[aI]
        if aH is not None and aH.getType() == IDrawData.TYPE_MESH:
            if isinstance(aH, Mesh):
                return aH.getIndexArray()

        return None

    @staticmethod
    def loadModel_exe(model, buf: bytes):

        if not (isinstance(buf, bytes)):
            raise RuntimeError("param error")

        br = BinaryReader(buf)
        magic1 = br.readByte()
        magic2 = br.readByte()
        magic3 = br.readByte()
        # magic = 'moc'
        if magic1 == 109 and magic2 == 111 and magic3 == 99:
            version = br.readByte()
        else:
            raise RuntimeError("Invalid MOC file.")

        br.setFormatVersion(version)
        if version > LIVE2D_FORMAT_VERSION_AVAILABLE:
            es = "Unsupported version %d\n" % version
            raise RuntimeError(es)

        aL = br.readObject()
        if version >= LIVE2D_FORMAT_VERSION_V2_8_TEX_OPTION:
            aH = br.readUShort()
            aT = br.readUShort()
            if aH != -30584 or aT != -30584:
                raise RuntimeError("Invalid load EOF")

        model.setModelImpl(aL)
        model_context = model.getModelContext()
        model_context.setDrawParam(model.getDrawParam())
        model_context.init()



from abc import ABC, abstractmethod

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .binary_reader import BinaryReader


class ISerializable(ABC):

    @abstractmethod
    def read(self, br: 'BinaryReader'):
        pass



from ...core import AMotion, Array
from ..Live2DFramework import Live2DFramework
from .l2d_expression_param import L2DExpressionParam


class L2DExpressionMotion(AMotion):
    EXPRESSION_DEFAULT = "DEFAULT"
    TYPE_SET = 0
    TYPE_ADD = 1
    TYPE_MULT = 2

    def __init__(self):
        super().__init__()
        self.paramList = Array()

    def updateParamExe(self, model, timeMSec, weight, motionQueueEnt):
        for i in range(len(self.paramList) - 1, -1, -1):
            param = self.paramList[i]
            if param.type == L2DExpressionMotion.TYPE_ADD:
                model.addToParamFloat(param.id, param.value, weight)
            elif param.type == L2DExpressionMotion.TYPE_MULT:
                model.multParamFloat(param.id, param.value, weight)
            elif param.type == L2DExpressionMotion.TYPE_SET:
                model.setParamFloat(param.id, param.value, weight)

    @staticmethod
    def loadJson(buf):
        ret = L2DExpressionMotion()
        pm = Live2DFramework.getPlatformManager()
        js = pm.jsonParseFromBytes(buf)
        ret.setFadeIn(int(js.get("fade_in", 0)) if int(js.get("fade_in", 0)) > 0 else 1000)
        ret.setFadeOut(int(js.get("fade_out", 0)) if int(js.get("fade_out", 0)) > 0 else 1000)
        if js.get("params", None) is None:
            return ret

        params = js["params"]
        param_num = len(params)
        ret.paramList = []
        for i in range(0, param_num, 1):
            param = params[i]
            param_id = str(param["id"])
            value = float(param["val"])
            calc = str(param.get("calc", "add"))
            if calc == "add":
                calc_type_int = L2DExpressionMotion.TYPE_ADD
            elif calc == "mult":
                calc_type_int = L2DExpressionMotion.TYPE_MULT
            elif calc == "set":
                calc_type_int = L2DExpressionMotion.TYPE_SET
            else:
                calc_type_int = L2DExpressionMotion.TYPE_ADD

            if calc_type_int == L2DExpressionMotion.TYPE_ADD:
                default_value = float(param.get("def", 0))
                value = value - default_value
            elif calc_type_int == L2DExpressionMotion.TYPE_MULT:
                default_value = float(param.get("def", 1))
                if default_value == 0:
                    default_value = 1
                value = value / default_value

            item = L2DExpressionParam()
            item.id = param_id
            item.type = calc_type_int
            item.value = value
            ret.paramList.append(item)

        return ret



from .deformer import Deformer
from .warp_context import WarpContext
from ..live2d import Live2D
from ..param import PivotManager
from ..type import Float32Array
from ..util import UtInterpolate


class WarpDeformer(Deformer):
    paramOutSide = [False]

    def __init__(self):
        super().__init__()
        self.row = 0
        self.col = 0
        self.pivotMgr = None
        self.pivotPoints = None

    def read(self, br):
        super().read(br)
        self.col = br.readInt32()
        self.row = br.readInt32()
        self.pivotMgr = br.readObject()
        self.pivotPoints = br.readObject()
        super().readOpacity(br)

    def init(self, mc):
        aI = WarpContext(self)
        aJ = (self.row + 1) * (self.col + 1)
        if aI.interpolatedPoints is not None:
            aI.interpolatedPoints = None

        aI.interpolatedPoints = Float32Array(aJ * 2)
        if aI.transformedPoints is not None:
            aI.transformedPoints = None

        if self.needTransform():
            aI.transformedPoints = Float32Array(aJ * 2)
        else:
            aI.transformedPoints = None

        return aI

    def setupInterpolate(self, modelContext, deformerContext: 'WarpContext'):
        aK = deformerContext
        if not self.pivotMgr.checkParamUpdated(modelContext):
            return

        aL = self.VT_()
        aH = WarpDeformer.paramOutSide
        aH[0] = False
        UtInterpolate.interpolatePoints(modelContext, self.pivotMgr, aH, aL, self.pivotPoints, aK.interpolatedPoints, 0,
                                        2)
        deformerContext.setOutsideParam(aH[0])
        self.interpolateOpacity(modelContext, self.pivotMgr, deformerContext, aH)

    def setupTransform(self, modelContext, deformerContext):
        aL = deformerContext
        aL.setAvailable(True)
        if not self.needTransform():
            aL.setTotalOpacity(aL.getInterpolatedOpacity())
        else:
            aH = self.getTargetId()
            if aL.tmpDeformerIndex == Deformer.DEFORMER_INDEX_NOT_INIT:
                aL.tmpDeformerIndex = modelContext.getDeformerIndex(aH)

            if aL.tmpDeformerIndex < 0:
                print("deformer is not reachable")

                aL.setAvailable(False)
            else:
                aN = modelContext.getDeformer(aL.tmpDeformerIndex)
                aI = modelContext.getDeformerContext(aL.tmpDeformerIndex)
                if aN is not None and aI.isAvailable():
                    aM = aI.getTotalScale()
                    aL.setTotalScale_notForClient(aM)
                    aO = aI.getTotalOpacity()
                    aL.setTotalOpacity(aO * aL.getInterpolatedOpacity())
                    aN.transformPoints(modelContext, aI, aL.interpolatedPoints, aL.transformedPoints, self.VT_(), 0, 2)
                    aL.setAvailable(True)
                else:
                    aL.setAvailable(False)

    def transformPoints(self, mc, dc: 'WarpContext', srcPoints, dstPoints, numPoint, ptOffset, ptStep):
        pivot_points = dc.transformedPoints if (dc.transformedPoints is not None) else dc.interpolatedPoints
        WarpDeformer.transformPoints_sdk2(srcPoints, dstPoints, numPoint, ptOffset, ptStep, pivot_points, self.row,
                                          self.col)

    def VT_(self):
        return (self.row + 1) * (self.col + 1)

    def getType(self):
        return Deformer.TYPE_WARP

    @staticmethod
    def transformPoints_sdk2(hvs, dst, pointCount, srcOffset, srcStep, grid, row, col):
        aW = pointCount * srcStep
        aT = 0
        aS = 0
        bl = 0
        bk = 0
        bf = 0
        be = 0
        aZ = False
        for ba in range(srcOffset, aW, srcStep):
            a4 = hvs[ba]
            aX = hvs[ba + 1]
            bd = a4 * row
            a7 = aX * col
            if bd < 0 or a7 < 0 or row <= bd or col <= a7:
                a1 = row + 1
                if not aZ:
                    aZ = True
                    aT = 0.25 * (grid[((0) + (0) * a1) * 2] + grid[((row) + (0) * a1) * 2] + grid[
                        ((0) + (col) * a1) * 2] +
                                 grid[((row) + (col) * a1) * 2])
                    aS = 0.25 * (grid[((0) + (0) * a1) * 2 + 1] + grid[((row) + (0) * a1) * 2 + 1] + grid[
                        ((0) + (col) * a1) * 2 + 1] + grid[((row) + (col) * a1) * 2 + 1])
                    aM = grid[((row) + (col) * a1) * 2] - grid[((0) + (0) * a1) * 2]
                    aL = grid[((row) + (col) * a1) * 2 + 1] - grid[((0) + (0) * a1) * 2 + 1]
                    bh = grid[((row) + (0) * a1) * 2] - grid[((0) + (col) * a1) * 2]
                    bg = grid[((row) + (0) * a1) * 2 + 1] - grid[((0) + (col) * a1) * 2 + 1]
                    bl = (aM + bh) * 0.5
                    bk = (aL + bg) * 0.5
                    bf = (aM - bh) * 0.5
                    be = (aL - bg) * 0.5
                    aT -= 0.5 * (bl + bf)
                    aS -= 0.5 * (bk + be)

                if (-2 < a4 and a4 < 3) and (-2 < aX and aX < 3):
                    if a4 <= 0:
                        if aX <= 0:
                            a3 = grid[((0) + (0) * a1) * 2]
                            a2 = grid[((0) + (0) * a1) * 2 + 1]
                            a8 = aT - 2 * bl
                            a6 = aS - 2 * bk
                            aK = aT - 2 * bf
                            aJ = aS - 2 * be
                            aO = aT - 2 * bl - 2 * bf
                            aN = aS - 2 * bk - 2 * be
                            bj = 0.5 * (a4 - (-2))
                            bi = 0.5 * (aX - (-2))
                            if bj + bi <= 1:
                                dst[ba] = aO + (aK - aO) * bj + (a8 - aO) * bi
                                dst[ba + 1] = aN + (aJ - aN) * bj + (a6 - aN) * bi
                            else:
                                dst[ba] = a3 + (a8 - a3) * (1 - bj) + (aK - a3) * (1 - bi)
                                dst[ba + 1] = a2 + (a6 - a2) * (1 - bj) + (aJ - a2) * (1 - bi)
                        else:
                            if aX >= 1:
                                aK = grid[((0) + (col) * a1) * 2]
                                aJ = grid[((0) + (col) * a1) * 2 + 1]
                                aO = aT - 2 * bl + 1 * bf
                                aN = aS - 2 * bk + 1 * be
                                a3 = aT + 3 * bf
                                a2 = aS + 3 * be
                                a8 = aT - 2 * bl + 3 * bf
                                a6 = aS - 2 * bk + 3 * be
                                bj = 0.5 * (a4 - (-2))
                                bi = 0.5 * (aX - (1))
                                if bj + bi <= 1:
                                    dst[ba] = aO + (aK - aO) * bj + (a8 - aO) * bi
                                    dst[ba + 1] = aN + (aJ - aN) * bj + (a6 - aN) * bi
                                else:
                                    dst[ba] = a3 + (a8 - a3) * (1 - bj) + (aK - a3) * (1 - bi)
                                    dst[ba + 1] = a2 + (a6 - a2) * (1 - bj) + (aJ - a2) * (1 - bi)
                            else:
                                aH = int(a7)
                                if aH == col:
                                    aH = col - 1

                                bj = 0.5 * (a4 - (-2))
                                bi = a7 - aH
                                bb = aH / col
                                a9 = (aH + 1) / col
                                aK = grid[((0) + (aH) * a1) * 2]
                                aJ = grid[((0) + (aH) * a1) * 2 + 1]
                                a3 = grid[((0) + (aH + 1) * a1) * 2]
                                a2 = grid[((0) + (aH + 1) * a1) * 2 + 1]
                                aO = aT - 2 * bl + bb * bf
                                aN = aS - 2 * bk + bb * be
                                a8 = aT - 2 * bl + a9 * bf
                                a6 = aS - 2 * bk + a9 * be
                                if bj + bi <= 1:
                                    dst[ba] = aO + (aK - aO) * bj + (a8 - aO) * bi
                                    dst[ba + 1] = aN + (aJ - aN) * bj + (a6 - aN) * bi
                                else:
                                    dst[ba] = a3 + (a8 - a3) * (1 - bj) + (aK - a3) * (1 - bi)
                                    dst[ba + 1] = a2 + (a6 - a2) * (1 - bj) + (aJ - a2) * (1 - bi)
                    else:
                        if 1 <= a4:
                            if aX <= 0:
                                a8 = grid[((row) + (0) * a1) * 2]
                                a6 = grid[((row) + (0) * a1) * 2 + 1]
                                a3 = aT + 3 * bl
                                a2 = aS + 3 * bk
                                aO = aT + 1 * bl - 2 * bf
                                aN = aS + 1 * bk - 2 * be
                                aK = aT + 3 * bl - 2 * bf
                                aJ = aS + 3 * bk - 2 * be
                                bj = 0.5 * (a4 - (1))
                                bi = 0.5 * (aX - (-2))
                                if bj + bi <= 1:
                                    dst[ba] = aO + (aK - aO) * bj + (a8 - aO) * bi
                                    dst[ba + 1] = aN + (aJ - aN) * bj + (a6 - aN) * bi
                                else:
                                    dst[ba] = a3 + (a8 - a3) * (1 - bj) + (aK - a3) * (1 - bi)
                                    dst[ba + 1] = a2 + (a6 - a2) * (1 - bj) + (aJ - a2) * (1 - bi)
                            else:
                                if aX >= 1:
                                    aO = grid[((row) + (col) * a1) * 2]
                                    aN = grid[((row) + (col) * a1) * 2 + 1]
                                    aK = aT + 3 * bl + 1 * bf
                                    aJ = aS + 3 * bk + 1 * be
                                    a8 = aT + 1 * bl + 3 * bf
                                    a6 = aS + 1 * bk + 3 * be
                                    a3 = aT + 3 * bl + 3 * bf
                                    a2 = aS + 3 * bk + 3 * be
                                    bj = 0.5 * (a4 - (1))
                                    bi = 0.5 * (aX - (1))
                                    if bj + bi <= 1:
                                        dst[ba] = aO + (aK - aO) * bj + (a8 - aO) * bi
                                        dst[ba + 1] = aN + (aJ - aN) * bj + (a6 - aN) * bi
                                    else:
                                        dst[ba] = a3 + (a8 - a3) * (1 - bj) + (aK - a3) * (1 - bi)
                                        dst[ba + 1] = a2 + (a6 - a2) * (1 - bj) + (aJ - a2) * (1 - bi)
                                else:
                                    aH = int(a7)
                                    if aH == col:
                                        aH = col - 1

                                    bj = 0.5 * (a4 - (1))
                                    bi = a7 - aH
                                    bb = aH / col
                                    a9 = (aH + 1) / col
                                    aO = grid[((row) + (aH) * a1) * 2]
                                    aN = grid[((row) + (aH) * a1) * 2 + 1]
                                    a8 = grid[((row) + (aH + 1) * a1) * 2]
                                    a6 = grid[((row) + (aH + 1) * a1) * 2 + 1]
                                    aK = aT + 3 * bl + bb * bf
                                    aJ = aS + 3 * bk + bb * be
                                    a3 = aT + 3 * bl + a9 * bf
                                    a2 = aS + 3 * bk + a9 * be
                                    if bj + bi <= 1:
                                        dst[ba] = aO + (aK - aO) * bj + (a8 - aO) * bi
                                        dst[ba + 1] = aN + (aJ - aN) * bj + (a6 - aN) * bi
                                    else:
                                        dst[ba] = a3 + (a8 - a3) * (1 - bj) + (aK - a3) * (1 - bi)
                                        dst[ba + 1] = a2 + (a6 - a2) * (1 - bj) + (aJ - a2) * (1 - bi)
                        else:
                            if aX <= 0:
                                aY = int(bd)
                                if aY == row:
                                    aY = row - 1

                                bj = bd - aY
                                bi = 0.5 * (aX - (-2))
                                bp = aY / row
                                bo = (aY + 1) / row
                                a8 = grid[((aY) + (0) * a1) * 2]
                                a6 = grid[((aY) + (0) * a1) * 2 + 1]
                                a3 = grid[((aY + 1) + (0) * a1) * 2]
                                a2 = grid[((aY + 1) + (0) * a1) * 2 + 1]
                                aO = aT + bp * bl - 2 * bf
                                aN = aS + bp * bk - 2 * be
                                aK = aT + bo * bl - 2 * bf
                                aJ = aS + bo * bk - 2 * be
                                if bj + bi <= 1:
                                    dst[ba] = aO + (aK - aO) * bj + (a8 - aO) * bi
                                    dst[ba + 1] = aN + (aJ - aN) * bj + (a6 - aN) * bi
                                else:
                                    dst[ba] = a3 + (a8 - a3) * (1 - bj) + (aK - a3) * (1 - bi)
                                    dst[ba + 1] = a2 + (a6 - a2) * (1 - bj) + (aJ - a2) * (1 - bi)
                            else:
                                if aX >= 1:
                                    aY = int(bd)
                                    if aY == row:
                                        aY = row - 1

                                    bj = bd - aY
                                    bi = 0.5 * (aX - (1))
                                    bp = aY / row
                                    bo = (aY + 1) / row
                                    aO = grid[((aY) + (col) * a1) * 2]
                                    aN = grid[((aY) + (col) * a1) * 2 + 1]
                                    aK = grid[((aY + 1) + (col) * a1) * 2]
                                    aJ = grid[((aY + 1) + (col) * a1) * 2 + 1]
                                    a8 = aT + bp * bl + 3 * bf
                                    a6 = aS + bp * bk + 3 * be
                                    a3 = aT + bo * bl + 3 * bf
                                    a2 = aS + bo * bk + 3 * be
                                    if bj + bi <= 1:
                                        dst[ba] = aO + (aK - aO) * bj + (a8 - aO) * bi
                                        dst[ba + 1] = aN + (aJ - aN) * bj + (a6 - aN) * bi
                                    else:
                                        dst[ba] = a3 + (a8 - a3) * (1 - bj) + (aK - a3) * (1 - bi)
                                        dst[ba + 1] = a2 + (a6 - a2) * (1 - bj) + (aJ - a2) * (1 - bi)
                                else:
                                    raise RuntimeError("error @BDBoxGrid")
                else:
                    dst[ba] = aT + a4 * bl + aX * bf
                    dst[ba + 1] = aS + a4 * bk + aX * be
            else:
                bn = bd - int(bd)
                bm = a7 - int(a7)
                aV = 2 * (int(bd) + int(a7) * (row + 1))
                if bn + bm < 1:
                    dst[ba] = grid[aV] * (1 - bn - bm) + grid[aV + 2] * bn + grid[aV + 2 * (row + 1)] * bm
                    dst[ba + 1] = grid[aV + 1] * (1 - bn - bm) + grid[aV + 3] * bn + grid[aV + 2 * (row + 1) + 1] * bm
                else:
                    dst[ba] = grid[aV + 2 * (row + 1) + 2] * (bn - 1 + bm) + grid[aV + 2 * (row + 1)] * (1 - bn) + grid[
                        aV + 2] * (1 - bm)
                    dst[ba + 1] = grid[aV + 2 * (row + 1) + 3] * (bn - 1 + bm) + grid[aV + 2 * (row + 1) + 1] * (
                            1 - bn) + \
                                  grid[aV + 3] * (1 - bm)



class MotionPriority:
    NONE = 0
    IDLE = 1
    NORMAL = 2
    FORCE = 3


class MotionGroup:
    IDLE = "idle"
    TAP_BODY = "tap_body"
    FLICK_HEAD = "flick_head"
    PINCH_IN = "pinch_in"
    PINCH_OUT = "pinch_out"
    SHAKE = "shake"


class HitArea:
    HEAD = "head"
    BODY = "body"



from .iphysics_param import IPhysicsParam
from .physics_constants import SRC_TO_Y, SRC_TO_G_ANGLE, SRC_TO_X


class PhysicsSrc(IPhysicsParam):

    def __init__(self, paramId, aK, scale, weight):
        super().__init__(aK, scale, weight)
        self.tL_ = None
        self.tL_ = paramId

    def update(self, aJ, aH):
        aK = self.scale * aJ.getParamFloat(self.paramId)
        aL = aH.getPhysicsPoint1()

        if self.tL_ == SRC_TO_X:
            aL.x = aL.x + (aK - aL.x) * self.weight
        elif self.tL_ == SRC_TO_Y:
            aL.y = aL.y + (aK - aL.y) * self.weight
        elif self.tL_ == SRC_TO_G_ANGLE:
            aI = aH.qr_()
            aI = aI + (aK - aI) * self.weight
            aH.pr_(aI)



import math
import os.path
from random import random, choice
from typing import TYPE_CHECKING, List, Union

from .core import UtSystem, log
from .framework import L2DBaseModel, L2DTargetPoint, L2DEyeBlink
from .lapp_define import MotionPriority
from .matrix_manager import MatrixManager
from .model_setting_json import ModelSettingJson
from .params import Parameter

if TYPE_CHECKING:
    from .core.draw import MeshContext, Mesh


class LAppModel(L2DBaseModel):

    def __init__(self):
        super().__init__()
        self.modelHomeDir = ""
        self.modelSetting = None
        self.matrixManager = MatrixManager()
        self.dragMgr = L2DTargetPoint()
        self.dragMgr.setPoint(0.0, 0.0)
        self.autoBreath = True
        self.autoBlink = True

        self.finishCallback = None

        self.__clearFlag = False

    def LoadModelJson(self, modelSettingPath: str):
        self.setUpdating(True)
        self.setInitialized(False)
        self.modelHomeDir = os.path.dirname(modelSettingPath) + "/"
        self.modelSetting = ModelSettingJson()

        self.modelSetting.loadModelSetting(modelSettingPath)

        path = self.modelHomeDir + self.modelSetting.getModelFile()

        self.loadModelData(path)

        for i in range(self.modelSetting.getTextureNum()):
            tex_paths = self.modelHomeDir + self.modelSetting.getTextureFile(i)
            self.loadTexture(i, tex_paths)

        if self.modelSetting.getExpressionNum() > 0:
            self.expressions = {}
            for j in range(self.modelSetting.getExpressionNum()):
                exp_name = self.modelSetting.getExpressionName(j)
                exp_file_path = self.modelHomeDir + self.modelSetting.getExpressionFile(j)
                self.loadExpression(exp_name, exp_file_path)
        else:
            self.expressionManager = None
            self.expressions = {}

        if self.eyeBlink is None:
            self.eyeBlink = L2DEyeBlink()

        if self.modelSetting.getPhysicsFile() is not None:
            self.loadPhysics(self.modelHomeDir + self.modelSetting.getPhysicsFile())
        else:
            self.physics = None

        if self.modelSetting.getPoseFile() is not None:
            pose = self.loadPose(self.modelHomeDir + self.modelSetting.getPoseFile())
            pose.updateParam(self.live2DModel)

        if self.modelSetting.getLayout() is not None:
            layout = self.modelSetting.getLayout()
            if layout.get("width", None) is not None:
                self.modelMatrix.setWidth(layout["width"])
            if layout.get("height", None) is not None:
                self.modelMatrix.setHeight(layout["height"])
            if layout.get("x", None) is not None:
                self.modelMatrix.setX(layout["x"])
            if layout.get("y", None) is not None:
                self.modelMatrix.setY(layout["y"])
            if layout.get("center_x", None) is not None:
                self.modelMatrix.centerX(layout["center_x"])
            if layout.get("center_y", None) is not None:
                self.modelMatrix.centerY(layout["center_y"])
            if layout.get("top", None) is not None:
                self.modelMatrix.top(layout["top"])
            if layout.get("bottom", None) is not None:
                self.modelMatrix.bottom(layout["bottom"])
            if layout.get("left", None) is not None:
                self.modelMatrix.left(layout["left"])
            if layout.get("right", None) is not None:
                self.modelMatrix.right(layout["right"])

        for j in range(self.modelSetting.getInitParamNum()):
            self.live2DModel.setParamFloat(self.modelSetting.getInitParamID(j),
                                           self.modelSetting.getInitParamValue(j))

        for j in range(self.modelSetting.getInitPartsVisibleNum()):
            self.live2DModel.setPartsOpacity(self.modelSetting.getInitPartsVisibleID(j),
                                             self.modelSetting.getInitPartsVisibleValue(j))

        self.live2DModel.saveParam()
        self.__preloadMotionGroup(MotionPriority.IDLE)
        self.mainMotionManager.stopAllMotions()
        self.setUpdating(False)
        self.setInitialized(True)

    def Resize(self, ww: int, wh: int):
        self.matrixManager.onResize(ww, wh)
        self.live2DModel.resize(ww, wh)

    def Drag(self, x: float, y: float):
        scx, scy = self.matrixManager.screenToScene(x, y)
        self.dragMgr.setPoint(scx, scy)

    def IsMotionFinished(self) -> bool:
        return self.mainMotionManager.isFinished()

    def SetOffset(self, dx: float, dy: float):
        self.matrixManager.setOffset(dx, dy)

    def SetScale(self, scale: float):
        self.matrixManager.setScale(scale)

    def SetParameterValue(self, paramId: str, value: float, weight: float = 1.0):
        self.live2DModel.setParamFloat(paramId, value, weight)

    def AddParameterValue(self, paramId: str, value: float, weight: float = 1.0):
        self.live2DModel.addToParamFloat(paramId, value, weight)

    def SetAutoBreathEnable(self, enable: bool):
        self.autoBreath = enable

    def SetAutoBlinkEnable(self, enable: bool):
        self.autoBlink = enable

    def GetParameterCount(self) -> int:
        return len(self.live2DModel.getModelContext().paramIdList)

    def GetParameter(self, index: int) -> Parameter:
        p = Parameter()
        p.value = self.live2DModel.getParamFloat(index)
        p.max = self.live2DModel.getModelContext().getParamMax(index)
        p.min = self.live2DModel.getModelContext().getParamMin(index)
        inner_params = self.live2DModel.getModelImpl().paramDefSet.getParamDefFloatList()
        is_inner = index < len(inner_params)
        p.type = Parameter.TYPE_INNER if is_inner else Parameter.TYPE_OUTER
        p.default = inner_params[index].defaultValue if is_inner else 0
        p.id = self.live2DModel.getModelContext().paramIdList[index]
        return p

    def GetPartCount(self) -> int:
        return len(self.live2DModel.getModelImpl().getPartsDataList())

    def GetPartId(self, index: int) -> str:
        return self.live2DModel.getModelContext().getPartsContext(index).partsData.id.id

    def GetPartIds(self) -> List[str]:
        return [str(i.id) for i in self.live2DModel.getModelContext().partsDataList]

    def SetPartOpacity(self, index: int, opacity: float):
        self.live2DModel.setPartsOpacity(index, opacity)

    def Update(self):
        self.dragMgr.update()
        self.setDrag(self.dragMgr.getX(), self.dragMgr.getY())

        time_m_sec = UtSystem.getUserTimeMSec() - self.startTimeMSec
        time_sec = time_m_sec / 1000.0
        t = time_sec * 2 * math.pi
        if self.mainMotionManager.isFinished():
            if callable(self.finishCallback):
                self.finishCallback()
                self.finishCallback = None

        updated = False
        if self.__clearFlag:
            self.mainMotionManager.stopAllMotions()
            if self.pose:
                self.pose.initParam(self.live2DModel)

            self.__clearFlag = False
        else:
            self.live2DModel.loadParam()
            updated = self.mainMotionManager.updateParam(self.live2DModel)
        self.live2DModel.saveParam()

        if not updated:
            if self.autoBlink and self.eyeBlink is not None:
                self.eyeBlink.updateParam(self.live2DModel)

        if self.expressionManager is not None and self.expressions is not None and not self.expressionManager.isFinished():
            self.expressionManager.updateParam(self.live2DModel)

        self.live2DModel.addToParamFloat("PARAM_ANGLE_X", self.dragX * 30, 1)
        self.live2DModel.addToParamFloat("PARAM_ANGLE_Y", self.dragY * 30, 1)
        self.live2DModel.addToParamFloat("PARAM_ANGLE_Z", (self.dragX * self.dragY) * -30, 1)
        self.live2DModel.addToParamFloat("PARAM_BODY_ANGLE_X", self.dragX * 10, 1)
        self.live2DModel.addToParamFloat("PARAM_EYE_BALL_X", self.dragX, 1)
        self.live2DModel.addToParamFloat("PARAM_EYE_BALL_Y", self.dragY, 1)

        if self.autoBreath:
            self.live2DModel.addToParamFloat("PARAM_ANGLE_X", float((15 * math.sin(t / 6.5345))), 0.5)
            self.live2DModel.addToParamFloat("PARAM_ANGLE_Y", float((8 * math.sin(t / 3.5345))), 0.5)
            self.live2DModel.addToParamFloat("PARAM_ANGLE_Z", float((10 * math.sin(t / 5.5345))), 0.5)
            self.live2DModel.addToParamFloat("PARAM_BODY_ANGLE_X", float((4 * math.sin(t / 15.5345))), 0.5)
            self.live2DModel.setParamFloat("PARAM_BREATH", float((0.5 + 0.5 * math.sin(t / 3.2345))), 1)

        if self.physics is not None:
            self.physics.updateParam(self.live2DModel)

        if self.pose is not None:
            self.pose.updateParam(self.live2DModel)

    def SetRandomExpression(self):
        tmp = []
        for name in self.expressions:
            tmp.append(name)

        no = int(random() * len(tmp))
        self.SetExpression(tmp[no])

    def StartRandomMotion(self, name=None, priority=MotionPriority.IDLE, onStartMotionHandler=None,
                          onFinishMotionHandler=None):
        if name is None:
            names = self.modelSetting.getMotionNames()
            if names is not None:
                name = choice(names)
            else:
                name = MotionPriority.IDLE
        count = self.modelSetting.getMotionNum(name)
        no = int(random() * count)
        self.StartMotion(name, no, priority, onStartMotionHandler, onFinishMotionHandler)

    def StartMotion(self, name, no, priority, onStartMotionHandler=None, onFinishMotionHandler=None):
        motion_name = self.modelSetting.getMotionFile(name, no)
        if motion_name is None or motion_name == "":
            if callable(onStartMotionHandler):
                onStartMotionHandler(name, no)
            if callable(onFinishMotionHandler):
                onFinishMotionHandler()
            return

        if priority == MotionPriority.FORCE:
            self.mainMotionManager.setReservePriority(priority)
        elif not self.mainMotionManager.reserveMotion(priority):
            return

        if self.motions.get(name) is None:
            mtn = self.loadMotion(None, self.modelHomeDir + motion_name)
        else:
            mtn = self.motions[name]

        self.finishCallback = onFinishMotionHandler
        if callable(onStartMotionHandler):
            onStartMotionHandler(name, no)
        self.__setFadeInFadeOut(name, no, priority, mtn)

    def SetExpression(self, name: str):
        motion = self.expressions[name]
        self.expressionManager.startMotion(motion, False)

    def Draw(self):
        # 根据设置的参数更新绘图所需参数
        self.live2DModel.update()
        model_matrix = self.modelMatrix
        tmp_matrix = self.matrixManager.getMvp(model_matrix)
        self.live2DModel.setMatrix(tmp_matrix)
        self.live2DModel.draw()

    def HitTest(self, hitAreaName: str, testX, testY) -> Union[str, None]:
        size = self.modelSetting.getHitAreaNum()
        for i in range(size):
            area_id = self.modelSetting.getHitAreaName(i)
            draw_id = self.modelSetting.getHitAreaID(i)
            if self.hitTestSimple(draw_id, testX, testY):
                return area_id

        return None

    def __preloadMotionGroup(self, name):
        for i in range(self.modelSetting.getMotionNum(name)):
            file = self.modelSetting.getMotionFile(name, i)
            motion = self.loadMotion(file, self.modelHomeDir + file)

            motion.setFadeIn(self.modelSetting.getMotionFadeIn(name, i))
            motion.setFadeOut(self.modelSetting.getMotionFadeOut(name, i))

    def __setFadeInFadeOut(self, name, no, priority, motion):
        motion.setFadeIn(self.modelSetting.getMotionFadeIn(name, no))
        motion.setFadeOut(self.modelSetting.getMotionFadeOut(name, no))

        self.mainMotionManager.startMotionPrio(motion, priority)

    def HitPart(self, src_x: float, src_y: float, topOnly: bool = False) -> List[str]:
        src_x, src_y = self.matrixManager.screenToScene(src_x, src_y)
        mx, my = self.matrixManager.invertTransform(src_x, src_y)
        mctx = self.live2DModel.getModelContext()
        draw_orders = mctx.nextList_drawIndex
        if not draw_orders:
            return []
        draw_orders = reversed(draw_orders)
        hit_part_ids = []
        for idx in draw_orders:
            if idx == -1:
                continue

            ddcxt: 'MeshContext' = mctx.getDrawContext(idx)
            pctx = mctx.getPartsContext(ddcxt.partsIndex)
            parent_part = pctx.partsData
            if not parent_part.visible or pctx.partsOpacity < 0.1:
                continue

            part_id = str(parent_part.id)
            if part_id in hit_part_ids:
                continue

            dd: 'Mesh' = ddcxt.drawData
            vertices = ddcxt.getTransformedPoints()
            indices = dd.indexArray
            size = len(indices)
            for i in range(0, size, 3):
                p1_idx = indices[i] * 2
                p2_idx = indices[i + 1] * 2
                p3_idx = indices[i + 2] * 2
                if not self.__isInTriangle(mx, my,
                                           vertices[p1_idx], vertices[p1_idx + 1],
                                           vertices[p2_idx], vertices[p2_idx + 1],
                                           vertices[p3_idx], vertices[p3_idx + 1]):
                    continue

                hit_part_ids.append(part_id)
                if topOnly:
                    return hit_part_ids
                break

        return hit_part_ids

    @staticmethod
    def __isInTriangle(x, y, x0, y0, x1, y1, x2, y2) -> bool:
        if x < min(x0, x1, x2):
            return False
        if x > max(x0, x1, x2):
            return False
        if y < min(y0, y1, y2):
            return False
        if y > max(y0, y1, y2):
            return False

        d_x = x - x2
        d_y = y - y2
        d_x21 = x2 - x1
        d_y12 = y1 - y2
        D = d_y12 * (x0 - x2) + d_x21 * (y0 - y2)
        s = d_y12 * d_x + d_x21 * d_y
        t = (y2 - y0) * d_x + (x0 - x2) * d_y
        if D < 0:
            return s <= 0 and t <= 0 and s + t >= D
        return s >= 0 and t >= 0 and s + t <= D

    def setPartScreenColor(self, part_index: int, r: float, g: float, b: float, a: float):
        self.live2DModel.modelContext.setPartScreenColor(part_index, r, g, b, a)

    def GetPartScreenColor(self, part_index: int) -> List[float]:
        """
        any modification to the returned list is equivalent to setPartScreenColor
        :param part_index:
        :return: ref of the part screen color
        """
        return self.live2DModel.modelContext.getPartScreenColor(part_index)

    def SetPartMultiplyColor(self, part_index: int, r: float, g: float, b: float, a: float):
        self.live2DModel.modelContext.setPartMultiplyColor(part_index, r, g, b, a)

    def GetPartMultiplyColor(self, part_index: int) -> List[float]:
        """
        modify the returned list is equivalent to setPartMultiplyColor
        :param part_index:
        :return: ref of the part multiply color
        """
        return self.live2DModel.modelContext.getPartMultiplyColor(part_index)

    def ClearMotions(self):
        self.__clearFlag = True
    
    def ResetExpression(self):
        self.expressionManager.stopAllMotions()


from ..type import Float32Array


class ClipMatrix:

    def __init__(self):
        self.m = Float32Array(16)
        self.identity()

    def identity(self):
        for aH in range(0, 16, 1):
            self.m[aH] = 1 if ((aH % 5) == 0) else 0

    def getArray(self):
        return self.m

    def getCopyMatrix(self):
        return Float32Array(len(self.m))

    def setMatrix(self, aI):
        if aI is None or len(aI) != 16:
            return

        for aH in range(0, 16, 1):
            self.m[aH] = aI[aH]

    def translate(self, aH, aJ, aI):
        self.m[12] = self.m[0] * aH + self.m[4] * aJ + self.m[8] * aI + self.m[12]
        self.m[13] = self.m[1] * aH + self.m[5] * aJ + self.m[9] * aI + self.m[13]
        self.m[14] = self.m[2] * aH + self.m[6] * aJ + self.m[10] * aI + self.m[14]
        self.m[15] = self.m[3] * aH + self.m[7] * aJ + self.m[11] * aI + self.m[15]

    def scale(self, aJ, aI, aH):
        self.m[0] *= aJ
        self.m[4] *= aI
        self.m[8] *= aH
        self.m[1] *= aJ
        self.m[5] *= aI
        self.m[9] *= aH
        self.m[2] *= aJ
        self.m[6] *= aI
        self.m[10] *= aH
        self.m[3] *= aJ
        self.m[7] *= aI
        self.m[11] *= aH



class StandardParams:
    ParamAngleX = "PARAM_ANGLE_X"
    ParamAngleY = "PARAM_ANGLE_Y"
    ParamAngleZ = "PARAM_ANGLE_Z"
    ParamEyeLOpen = "PARAM_EYE_L_OPEN"
    ParamEyeROpen = "PARAM_EYE_R_OPEN"
    ParamEyeLSmile = "PARAM_EYE_L_SMILE"
    ParamEyeRSmile = "PARAM_EYE_R_SMILE"
    ParamEyeBallX = "PARAM_EYE_BALL_X"
    ParamEyeBallY = "PARAM_EYE_BALL_Y"
    ParamEyeBallForm = "PARAM_EYE_BALL_FORM"
    ParamBrowLX = "PARAM_BROW_L_X"
    ParamBrowLY = "PARAM_BROW_L_Y"
    ParamBrowLAngle = "PARAM_BROW_L_ANGLE"
    ParamBrowLForm = "PARAM_BROW_L_FORM"
    ParamBrowRX = "PARAM_BROW_R_X"
    ParamBrowRY = "PARAM_BROW_R_Y"
    ParamBrowRAngle = "PARAM_BROW_R_ANGLE"
    ParamBrowRForm = "PARAM_BROW_R_FORM"
    ParamMouthOpenY = "PARAM_MOUTH_OPEN_Y"
    ParamMouthForm = "PARAM_MOUTH_FORM"
    ParamSmile = "PARAM_SMILE"
    ParamTere = "PARAM_TERE"
    ParamBodyAngleX = "PARAM_BODY_ANGLE_X"
    ParamBodyAngleZ = "PARAM_BODY_ANGLE_Z"
    ParamBreath = "PARAM_BREATH"
    ParamHairFront = "PARAM_HAIR_FRONT"
    ParamHairSide = "PARAM_HAIR_SIDE"
    ParamHairBack = "PARAM_HAIR_BACK"
    ParamHairFuwa = "PARAM_HAIR_FUWA"
    ParamShoulderX = "PARAM_SHOULDER_X"
    ParamBustX = "PARAM_BUST_X"
    ParamBustY = "PARAM_BUST_Y"
    ParamBaseX = "PARAM_BASE_X"
    ParamBaseY = "PARAM_BASE_Y"


class Parameter:
    TYPE_INNER = 0  # params defined in moc file
    TYPE_OUTER = 1  # params generated from motion file

    def __init__(self):
        self.id: str = ""
        self.type: int = 0
        self.value: float = 0
        self.max: float = 0
        self.min: float = 0
        self.default: float = 0

    def __str__(self):
        return f"Parameter(id={self.id}, type={self.type}, value={self.value}, max={self.max}, min={self.min}, default={self.default})"



import math


class UtMotion:
    @staticmethod
    def getEasingSine(value: float):
        if value < 0:
            return 0

        if value > 1:
            return 1

        return 0.5 - 0.5 * math.cos(value * math.pi)



from abc import ABC, abstractmethod

from ..type import Float32Array


class DrawParam(ABC):
    DEFAULT_FIXED_TEXTURE_COUNT = 32
    CLIPPING_PROCESS_NONE = 0
    CLIPPING_PROCESS_OVERWRITE_ALPHA = 1
    CLIPPING_PROCESS_MULTIPLY_ALPHA = 2
    CLIPPING_PROCESS_DRAW = 3
    CLIPPING_PROCESS_CLEAR_ALPHA = 4

    def __init__(self):
        self.fixedTextureCount = DrawParam.DEFAULT_FIXED_TEXTURE_COUNT
        self.baseAlpha = 1
        self.baseRed = 1
        self.baseGreen = 1
        self.baseBlue = 1
        self.culling = False
        self.matrix4x4 = Float32Array(16)
        self.preMultipliedAlpha = False
        self.anisotropy = 0
        self.clippingProcess = DrawParam.CLIPPING_PROCESS_NONE
        self.clipBufPre_clipContextMask = None
        self.clipBufPre_clipContextDraw = None
        self.channel_colors = {}

    def setChannelFlagAsColor(self, aH, aI):
        self.channel_colors[aH] = aI

    def getChannelFlagAsColor(self, aY):
        return self.channel_colors[aY]

    @abstractmethod
    def setupDraw(self):
        pass

    @abstractmethod
    def drawTexture(self, texNo, screenColor, indexArray, vertexArray, uvArray, opacity, comp, multiplyColor):
        pass

    def setCulling(self, aH):
        self.culling = aH

    def setMatrix(self, aH):
        for aI in range(0, 16, 1):
            self.matrix4x4[aI] = aH[aI]

    def getMatrix(self):
        return self.matrix4x4

    def setPreMultipliedAlpha(self, aH):
        self.preMultipliedAlpha = aH

    def isPreMultipliedAlpha(self):
        return self.preMultipliedAlpha

    def setAnisotropy(self, aH):
        self.anisotropy = aH

    def getAnisotropy(self):
        return self.anisotropy

    def getClippingProcess(self):
        return self.clippingProcess

    def setClippingProcess(self, aH):
        self.clippingProcess = aH

    def setClipBufPre_clipContextForMask(self, aH):
        self.clipBufPre_clipContextMask = aH

    def getClipBufPre_clipContextMask(self):
        return self.clipBufPre_clipContextMask

    def setClipBufPre_clipContextForDraw(self, aH):
        self.clipBufPre_clipContextDraw = aH

    def getClipBufPre_clipContextDraw(self):
        return self.clipBufPre_clipContextDraw



from typing import TYPE_CHECKING

from .idraw_data import IDrawData
from .mesh_context import MeshContext
from ..DEF import LIVE2D_FORMAT_VERSION_V2_8_TEX_OPTION, VERTEX_STEP, VERTEX_TYPE, VERTEX_OFFSET, \
    VERTEX_TYPE_OFFSET0_STEP2, REVERSE_TEXTURE_T, VERTEX_TYPE_OFFSET2_STEP5
from ..live2d import Live2D
from ..param import PivotManager
from ..type import Int16Array, Float32Array
from ..util import UtInterpolate

if TYPE_CHECKING:
    from ..model_context import ModelContext


class Mesh(IDrawData):
    INSTANCE_COUNT = 0
    MASK_COLOR_COMPOSITION = 30
    COLOR_COMPOSITION_NORMAL = 0
    COLOR_COMPOSITION_SCREEN = 1
    COLOR_COMPOSITION_MULTIPLY = 2
    paramOutside = [False]

    def __init__(self):
        super().__init__()
        self.textureNo = -1
        self.pointCount = 0
        self.polygonCount = 0
        self.optionFlag = None
        self.indexArray = None
        self.pivotPoints = None
        self.uvs = None
        self.colorCompositionType = Mesh.COLOR_COMPOSITION_NORMAL
        self.culling = True
        self.instanceNo = Mesh.INSTANCE_COUNT
        Mesh.INSTANCE_COUNT += 1

    def setTextureNo(self, aH):
        self.textureNo = aH

    def getTextureNo(self):
        return self.textureNo

    def getUvs(self):
        return self.uvs

    def getOptionFlag(self):
        return self.optionFlag

    def getNumPoints(self):
        return self.pointCount

    def getType(self):
        return IDrawData.TYPE_MESH

    def read(self, br):
        super().read(br)
        self.textureNo = br.readInt32()
        self.pointCount = br.readInt32()
        self.polygonCount = br.readInt32()
        obj = br.readObject()
        self.indexArray = Int16Array(self.polygonCount * 3)
        for aJ in range(self.polygonCount * 3 - 1, 0 - 1, -1):
            self.indexArray[aJ] = obj[aJ]

        self.pivotPoints = br.readObject()
        self.uvs = br.readObject()
        if br.getFormatVersion() >= LIVE2D_FORMAT_VERSION_V2_8_TEX_OPTION:
            self.optionFlag = br.readInt32()
            if self.optionFlag != 0:
                if (self.optionFlag & 1) != 0:
                    _ = br.readInt32()
                    raise RuntimeError("not handled")

                if (self.optionFlag & Mesh.MASK_COLOR_COMPOSITION) != 0:
                    self.colorCompositionType = (self.optionFlag & Mesh.MASK_COLOR_COMPOSITION) >> 1
                else:
                    self.colorCompositionType = Mesh.COLOR_COMPOSITION_NORMAL

                if (self.optionFlag & 32) != 0:
                    self.culling = False
        else:
            self.optionFlag = 0

    def init(self, aL):
        ctx = MeshContext(self)
        aI = self.pointCount * VERTEX_STEP
        aH = self.needTransform()
        if ctx.interpolatedPoints is not None:
            ctx.interpolatedPoints = None

        ctx.interpolatedPoints = Float32Array(aI)
        if ctx.transformedPoints is not None:
            ctx.transformedPoints = None

        ctx.transformedPoints = Float32Array(aI) if aH else None
        aM = VERTEX_TYPE

        if aM == VERTEX_TYPE_OFFSET0_STEP2:
            if REVERSE_TEXTURE_T:
                for aJ in range(self.pointCount - 1, 0 - 1, -1):
                    aO = aJ << 1
                    self.uvs[aO + 1] = 1 - self.uvs[aO + 1]
        elif aM == VERTEX_TYPE_OFFSET2_STEP5:
            for aJ in range(self.pointCount - 1, 0 - 1, -1):
                aO = aJ << 1
                aK = aJ * VERTEX_STEP
                aQ = self.uvs[aO]
                aP = self.uvs[aO + 1]
                ctx.interpolatedPoints[aK] = aQ
                ctx.interpolatedPoints[aK + 1] = aP
                ctx.interpolatedPoints[aK + 4] = 0
                if aH:
                    ctx.transformedPoints[aK] = aQ
                    ctx.transformedPoints[aK + 1] = aP
                    ctx.transformedPoints[aK + 4] = 0

        return ctx

    def setupInterpolate(self, aJ, aH):
        aK = aH
        if not (self == aK.getDrawData()):
            print("### assert!! ### ")

        if not self.pivotMgr.checkParamUpdated(aJ):
            return

        super().setupInterpolate(aJ, aK)
        if aK.paramOutside:
            return

        aI = Mesh.paramOutside
        aI[0] = False
        UtInterpolate.interpolatePoints(aJ, self.pivotMgr, aI, self.pointCount, self.pivotPoints, aK.interpolatedPoints,
                                        VERTEX_OFFSET, VERTEX_STEP)

    def setupTransform(self, mc, dc=None):
        if not (self == dc.getDrawData()):
            raise RuntimeError("context not match")

        aL = False
        if dc.paramOutside:
            aL = True

        if not aL:
            super().setupTransform(mc)
            if self.needTransform():
                target_id = self.getTargetId()
                if dc.tmpDeformerIndex == IDrawData.DEFORMER_INDEX_NOT_INIT:
                    dc.tmpDeformerIndex = mc.getDeformerIndex(target_id)

                if dc.tmpDeformerIndex < 0:
                    print(f"deformer not found: {target_id}")
                else:
                    d = mc.getDeformer(dc.tmpDeformerIndex)
                    dctx = mc.getDeformerContext(dc.tmpDeformerIndex)
                    if d is not None and not dctx.isOutsideParam():
                        d.transformPoints(mc, dctx, dc.interpolatedPoints, dc.transformedPoints, self.pointCount,
                                          VERTEX_OFFSET, VERTEX_STEP)
                        dc.available = True
                    else:
                        dc.available = False

                    dc.baseOpacity = dctx.getTotalOpacity()

    def draw(self, dp, mctx: 'ModelContext', dctx: 'MeshContext'):
        if not (self == dctx.getDrawData()):
            raise RuntimeError("context not match")

        if dctx.paramOutside:
            return

        texNr = self.textureNo
        if texNr < 0:
            texNr = 1

        opacity = self.getOpacity(dctx) * dctx.partsOpacity * dctx.baseOpacity
        vertices = dctx.transformedPoints if (dctx.transformedPoints is not None) else dctx.interpolatedPoints
        dp.setClipBufPre_clipContextForDraw(dctx.clipBufPre_clipContext)
        dp.setCulling(self.culling)
        pctx = mctx.getPartsContext(dctx.partsIndex)
        dp.drawTexture(texNr, pctx.screenColor, self.indexArray, vertices, self.uvs, opacity, self.colorCompositionType,
                       pctx.multiplyColor)

    def getIndexArray(self):
        return self.indexArray



from .alive2d_model import ALive2DModel
from .graphics import DrawParamOpenGL


class Live2DModelOpenGL(ALive2DModel):

    def __init__(self):
        super().__init__()
        self.drawParamGL: DrawParamOpenGL = DrawParamOpenGL()

    def resize(self, ww: int, wh: int):
        self.drawParamGL.resize(ww, wh)

    def update(self):
        self.modelContext.update()
        self.modelContext.preDraw(self.drawParamGL)

    def draw(self):
        self.modelContext.draw(self.drawParamGL)

    def getDrawParam(self):
        return self.drawParamGL

    def setMatrix(self, aH):
        self.drawParamGL.setMatrix(aH)

    @staticmethod
    def loadModel(aI):
        aH = Live2DModelOpenGL()
        ALive2DModel.loadModel_exe(aH, aI)
        return aH

    def setTexture(self, aI, aH):
        if self.drawParamGL is None:
            raise RuntimeError("current gl is none")

        self.drawParamGL.setTexture(aI, aH)



from typing import Union, List


class UtString:
    @staticmethod
    def startswith(s: bytes, offset: int, pat: str) -> bool:
        end_pos = offset + len(pat)
        if end_pos >= len(s):
            return False

        for i in range(offset, end_pos, 1):
            if chr(s[i]) != pat[i - offset]:
                return False

        return True

    @staticmethod
    def createString(buf: bytes, offset: int, size: int) -> str:
        return buf[offset:offset + size].decode("utf-8")

    @staticmethod
    def strToFloat(s: bytes, length: int, offset: int, ret: List[bool]) -> float:
        result = 0
        _n = 10
        _p = False
        neg = chr(s[offset]) == '-'
        if neg:
            offset += 1
        while offset < length:
            c = chr(s[offset])
            if '0' <= c <= '9':
                if _p:
                    result += float(c) / _n
                    _n *= 10
                else:
                    result = result * 10 + float(c)
            elif c == '.':
                _p = True
            else:
                break

            offset += 1
        if neg:
            result = -result
        ret[0] = offset
        return result



import struct
from typing import List, Any

from .live2d_object_factory import Live2DObjectFactory
from ..DEF import OBJECT_REF
from ..id import Id
from ..type import Int32Array, Float32Array, Float64Array, Array


class BinaryReader:

    def __init__(self, buf: bytes):
        self.offset8Bit = 0
        self.current8Bit = 0
        self.formatVersion = 0
        self.objects: List[Any] = []
        self.buf = buf
        self.offset = 0

    def readNumber(self):
        b1 = self.readByte()
        if (b1 & 128) == 0:
            return b1 & 255

        b2 = self.readByte()
        if (b2 & 128) == 0:
            return ((b1 & 127) << 7) | (b2 & 127)

        b3 = self.readByte()
        if (b3 & 128) == 0:
            return ((b1 & 127) << 14) | ((b2 & 127) << 7) | (b3 & 255)

        b4 = self.readByte()
        if (b4 & 128) == 0:
            return ((b1 & 127) << 21) | ((b2 & 127) << 14) | ((b3 & 127) << 7) | (b4 & 255)

        raise RuntimeError("number parse error")

    def getFormatVersion(self):
        return self.formatVersion

    def setFormatVersion(self, aH):
        self.formatVersion = aH

    def readType(self):
        return self.readNumber()

    def readDouble(self):
        self.checkBits()
        ret = self.offset
        self.offset += 8
        return struct.unpack('>d', self.buf[ret:ret + 8])[0]

    def readFloat32(self):
        self.checkBits()
        ret = self.offset
        self.offset += 4
        return struct.unpack('>f', self.buf[ret:ret + 4])[0]

    def readInt32(self):
        self.checkBits()
        ret = self.offset
        self.offset += 4
        return struct.unpack('>i', self.buf[ret:ret + 4])[0]

    def readByte(self):
        self.checkBits()
        ret = self.offset
        self.offset += 1
        return self.buf[ret]

    def readUShort(self):
        self.checkBits()
        ret = self.offset
        self.offset += 2
        return struct.unpack('>h', self.buf[ret:ret + 2])[0]

    def readLong(self):
        self.checkBits()
        self.offset += 8
        raise RuntimeError("_L _q read long")

    def readBoolean(self):
        self.checkBits()
        ret = self.offset
        self.offset += 1
        return self.buf[ret] != 0

    def readUTF8String(self):
        self.checkBits()
        aH = self.readType()
        result = self.buf[self.offset:self.offset + aH].decode("utf-8")
        self.offset += aH
        return result

    def readInt32Array(self):
        self.checkBits()
        aI = self.readType()
        aH = Int32Array(aI)
        for aJ in range(0, aI, 1):
            aH[aJ] = self.readInt32()

        return aH

    def readFloat32Array(self):
        self.checkBits()
        aI = self.readType()
        aH = Float32Array(aI)
        for aJ in range(0, aI, 1):
            aH[aJ] = self.readFloat32()

        return aH

    def readFloat64Array(self):
        self.checkBits()
        aI = self.readType()
        aH = Float64Array(aI)
        for aJ in range(0, aI, 1):
            aH[aJ] = self.readDouble()

        return aH


    def readObject(self, aJ = -1):
        self.checkBits()
        if aJ < 0:
            aJ = self.readType()

        if aJ == OBJECT_REF:
            aH = self.readInt32()
            if 0 <= aH < len(self.objects):
                return self.objects[aH]
            else:
                raise RuntimeError("_sL _4i @_m0")
        else:
            aI = self.readKnownTypeObject(aJ)
            self.objects.append(aI)
            return aI

    def readKnownTypeObject(self, aN):
        if aN == 0:
            return None
        elif aN == 50:
            aK = self.readUTF8String()
            aI = Id.getID(aK)
            return aI
        elif aN == 51:
            aK = self.readUTF8String()
            aI = Id.getID(aK)
            return aI
        elif aN == 134:
            aK = self.readUTF8String()
            aI = Id.getID(aK)
            return aI
        elif aN == 60:
            aK = self.readUTF8String()
            aI = Id.getID(aK)
            return aI
        elif aN >= 48:
            aL = Live2DObjectFactory.create(aN)
            aL.read(self)
            return aL
        elif aN == 1:
            return self.readUTF8String()
        elif aN == 15:
            aH = self.readType()
            aI = Array(aH)
            for aJ in range(0, aH, 1):
                aI[aJ] = self.readObject()
            return aI
        elif aN == 23:
            raise RuntimeError("type not implemented")
        elif aN == 16 or aN == 25:
            return self.readInt32Array()
        elif aN == 26:
            return self.readFloat64Array()
        elif aN == 27:
            return self.readFloat32Array()
        raise RuntimeError("type error %d" % aN)

    def readBit(self):
        if self.offset8Bit == 0:
            self.current8Bit = self.readByte()
        elif self.offset8Bit == 8:
            self.current8Bit = self.readByte()
            self.offset8Bit = 0

        ret = ((self.current8Bit >> (7 - self.offset8Bit)) & 1) == 1
        self.offset8Bit += 1
        return ret

    def checkBits(self):
        if self.offset8Bit != 0:
            self.offset8Bit = 0



from .deformer import Deformer
from .roation_deformer import RotationDeformer, AffineEnt
from .warp_deformer import WarpDeformer
from .rotation_context import RotationContext
from .warp_context import WarpContext
from .deformer_context import DeformerContext



from ..io.iserializable import ISerializable
from ..type import Array


class ParamDefSet(ISerializable):

    def __init__(self):
        self.paramDefList = None

    def getParamDefFloatList(self):
        return self.paramDefList

    def read(self, br):
        self.paramDefList = br.readObject()



from abc import abstractmethod
from typing import TYPE_CHECKING, List, Optional

from ..DEF import LIVE2D_FORMAT_VERSION_V2_10_SDK2
from ..id import Id
from ..io.iserializable import ISerializable
from ..util import UtInterpolate

if TYPE_CHECKING:
    from ..io import BinaryReader
    from ..model_context import ModelContext
    from .deformer_context import DeformerContext
    from ..param import PivotManager



class Deformer(ISerializable):
    DEFORMER_INDEX_NOT_INIT = -2
    TYPE_ROTATION = 1
    TYPE_WARP = 2

    def __init__(self):
        self.id: Optional[Id] = None
        self.targetId: Optional[Id] = None
        self.dirty: bool = True
        self.pivotOpacities: Optional[List[float]] = None

    def read(self, br: 'BinaryReader'):
        self.id = br.readObject()
        self.targetId = br.readObject()

    def readOpacity(self, br: 'BinaryReader'):
        if br.getFormatVersion() >= LIVE2D_FORMAT_VERSION_V2_10_SDK2:
            self.pivotOpacities = br.readFloat32Array()

    @abstractmethod
    def init(self, mc: 'ModelContext'):
        pass

    @abstractmethod
    def setupInterpolate(self, modelContext: 'ModelContext', deformerContext: 'DeformerContext'):
        pass

    def interpolateOpacity(self, mdc, pivotMgr: 'PivotManager', bctx: 'DeformerContext', ret: List[bool]):
        if self.pivotOpacities is None:
            bctx.setInterpolatedOpacity(1)
        else:
            bctx.setInterpolatedOpacity(UtInterpolate.interpolateFloat(mdc, pivotMgr, ret, self.pivotOpacities))

    @abstractmethod
    def setupTransform(self, mc, dc) -> bool:
        pass

    @abstractmethod
    def transformPoints(self, mc: 'ModelContext', dc: 'DeformerContext',
                        srcPoints: List[float],
                        dstPoints: List[float],
                        numPoint: int,
                        ptOffset: int,
                        ptStep: int):
        pass

    @abstractmethod
    def getType(self) -> int:
        pass

    def setTargetId(self, aH):
        self.targetId = aH

    def setId(self, aH):
        self.id = aH

    def getTargetId(self) -> Id:
        return self.targetId

    def getId(self) -> Id:
        return self.id

    def needTransform(self) -> bool:
        return self.targetId is not None and (self.targetId != Id.DST_BASE_ID())



from .physics_constants import TARGET_FROM_ANGLE, TARGET_FROM_ANGLE_V
from .iphysics_param import IPhysicsParam


class PhysicsTarget(IPhysicsParam):

    def __init__(self, scale, aK, paramId, weight):
        super().__init__(aK, paramId, weight)
        self.YP_ = scale

    def update(self, aI, aH):
        if self.YP_ == TARGET_FROM_ANGLE:
            aI.setParamFloat(self.paramId, self.scale * aH._5r(), self.weight)
        elif self.YP_ == TARGET_FROM_ANGLE_V:
            aI.setParamFloat(self.paramId, self.scale * aH.Cs_(), self.weight)



from ..io.iserializable import ISerializable

class ParamDefFloat(ISerializable):

    def __init__(self):
        self.minValue = None
        self.maxValue = None
        self.defaultValue = None
        self.paramId = None

    def read(self, br):
        self.minValue = br.readFloat32()
        self.maxValue = br.readFloat32()
        self.defaultValue = br.readFloat32()
        self.paramId = br.readObject()

    def getMinValue(self) -> float:
        return self.minValue

    def getMaxValue(self) -> float:
        return self.maxValue

    def getDefaultValue(self) -> float:
        return self.defaultValue

    def getParamID(self):
        return self.paramId



VERTEX_TYPE_OFFSET0_STEP2: int = 1
VERTEX_TYPE_OFFSET2_STEP5: int = 2
VERTEX_OFFSET: int = 0
VERTEX_STEP: int = 2
VERTEX_TYPE: int = VERTEX_TYPE_OFFSET0_STEP2
REVERSE_TEXTURE_T: bool = True
MAX_INTERPOLATION: int = 5
PIVOT_TABLE_SIZE: int = 65
GOSA: float = 0.0001

LIVE2D_FORMAT_VERSION_V2_8_TEX_OPTION: int = 8
LIVE2D_FORMAT_VERSION_V2_10_SDK2: int = 10
LIVE2D_FORMAT_VERSION_V2_11_SDK2_1: int = 11
LIVE2D_FORMAT_VERSION_AVAILABLE: int = LIVE2D_FORMAT_VERSION_V2_11_SDK2_1
OBJECT_REF = 33



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..id import Id

class ClipDrawContext:

    def __init__(self, aI: 'Id', aH: int):
        self.drawDataId = aI
        self.drawDataIndex = aH



class PhysicsPoint:

    def __init__(self):
        self.mass = 1
        self.x = 0
        self.y = 0
        self.vx = 0
        self.vy = 0
        self.ax = 0
        self.ay = 0
        self.fx = 0
        self.fy = 0
        self.lastX = 0
        self.lastY = 0
        self.lastVX = 0
        self.lastVY = 0

    def setupLast(self):
        self.lastX = self.x
        self.lastY = self.y
        self.lastVX = self.vx
        self.lastVY = self.vy



import wave
import numpy as np
import time


class WavHandler:
    def __init__(self):
        # 每个通道的采样帧数
        self.numFrames: int = 0
        # 采样率，帧/秒
        self.sampleRate: int = 0
        self.sampleWidth: int = 0
        # 通道数
        self.numChannels: int = 0
        # 数据
        self.pcmData: np.ndarray = None
        # 已经读取的帧数
        self.lastOffset: int = 0
        # 当前rms值
        self.currentRms: float = 0
        # 开始读取的时间
        self.startTime: float = -1

    def Start(self, filePath: str) -> None:
        self.ReleasePcmData()
        try:
            with wave.open(filePath, "r") as wav:
                self.numFrames = wav.getnframes()
                self.sampleRate = wav.getframerate()
                self.sampleWidth = wav.getsampwidth()
                self.numChannels = wav.getnchannels()
                # 双声道 / 单声道
                self.pcmData = np.frombuffer(wav.readframes(self.numFrames), dtype=np.int16)
                # 标准化
                self.pcmData = self.pcmData / np.max(np.abs(self.pcmData))
                # 拆分通道
                self.pcmData = self.pcmData.reshape(-1, self.numChannels).T

                self.startTime = time.time()
                self.lastOffset = 0

        except Exception as e:
            self.ReleasePcmData()

    def ReleasePcmData(self):
        if self.pcmData is not None:
            del self.pcmData
            self.pcmData = None

    def GetRms(self) -> float:
        """
        获取当前音频响度
        """
        return self.currentRms

    def Update(self) -> bool:
        """
        更新位置
        """
        # 数据未加载或者数据已经读取完毕
        if self.pcmData is None or self.lastOffset >= self.numFrames:
            return False

        currentTime = time.time() - self.startTime
        currentOffset = int(currentTime * self.sampleRate)

        # 时间太短
        if currentOffset == self.lastOffset:
            return True

        currentOffset = min(currentOffset, self.numFrames)

        dataFragment = self.pcmData[:, self.lastOffset:currentOffset].astype(np.float32)

        self.currentRms = np.sqrt(np.mean(np.square(dataFragment)))

        self.lastOffset = currentOffset
        return True


from .idraw_context import IDrawContext
from .idraw_data import IDrawData


class MeshContext(IDrawContext):

    def __init__(self, dd):
        super().__init__(dd)
        self.tmpDeformerIndex = IDrawData.DEFORMER_INDEX_NOT_INIT
        self.interpolatedPoints = None
        self.transformedPoints = None

    def getTransformedPoints(self):
        return self.transformedPoints if (self.transformedPoints is not None) else self.interpolatedPoints



from typing import Dict, Union, Optional

from ...core import Live2DModelOpenGL, Live2DMotion
from ..Live2DFramework import Live2DFramework
from ..matrix import L2DModelMatrix
from ..motion import L2DExpressionMotion, L2DMotionManager
from ..physics import L2DPhysics
from ..pose import L2DPose


class L2DBaseModel:
    texCount = 0

    def __init__(self):
        self.live2DModel: Optional[Live2DModelOpenGL] = None
        self.modelMatrix: Optional[L2DModelMatrix] = None
        self.eyeBlink = None
        self.physics = None
        self.pose: Union[None, L2DPose] = None
        self.debugMode = False
        self.initialized = False
        self.updating = False
        self.alpha = 1
        self.accAlpha = 0
        self.accelX = 0
        self.accelY = 0
        self.accelZ = 0
        self.dragX = 0
        self.dragY = 0
        self.startTimeMSec = 0
        self.mainMotionManager = L2DMotionManager()
        self.expressionManager = L2DMotionManager()
        self.motions = {}
        self.expressions: Dict = {}
        self.isTexLoaded = False

    def getModelMatrix(self):
        return self.modelMatrix

    def setAlpha(self, a):
        if a > 0.999:
            a = 1
        if a < 0.001:
            a = 0
        self.alpha = a

    def getAlpha(self):
        return self.alpha

    def isInitialized(self):
        return self.initialized

    def setInitialized(self, v):
        self.initialized = v

    def isUpdating(self):
        return self.updating

    def setUpdating(self, v):
        self.updating = v

    def getLive2DModel(self):
        return self.live2DModel

    def setLipSync(self, v):
        self.lipSync = v

    def setLipSyncValue(self, v):
        self.lipSyncValue = v

    def setAccel(self, x, y, z):
        self.accelX = x
        self.accelY = y
        self.accelZ = z

    def setDrag(self, x, y):
        self.dragX = x
        self.dragY = y

    def getMainMotionManager(self):
        return self.mainMotionManager

    def getExpressionManager(self):
        return self.expressionManager

    def loadModelData(self, path):
        pm = Live2DFramework.getPlatformManager()
        if self.debugMode:
            pm.log("Load model : " + path)

        self.live2DModel = pm.loadLive2DModel(path)
        self.live2DModel.saveParam()

        self.modelMatrix = L2DModelMatrix(self.live2DModel.getCanvasWidth(),
                                          self.live2DModel.getCanvasHeight())
        self.modelMatrix.setWidth(2)
        self.modelMatrix.setCenterPosition(0, 0)

        return self.live2DModel

    def loadTexture(self, no, path):
        self.texCount += 1
        pm = Live2DFramework.getPlatformManager()
        if self.debugMode:
            pm.log("Load Texture : " + path)

        pm.loadTexture(self.live2DModel, no, path)

        self.texCount -= 1
        if self.texCount == 0:
            self.isTexLoaded = True

    def loadMotion(self, name, path):
        pm = Live2DFramework.getPlatformManager()
        if self.debugMode:
            pm.log("Load Motion : " + path)

        buf = pm.loadBytes(path)

        motion = Live2DMotion.loadMotion(buf)
        if name is not None:
            self.motions[name] = motion
        return motion

    def loadExpression(self, name, path):
        pm = Live2DFramework.getPlatformManager()
        if self.debugMode:
            pm.log("Load Expression : " + path)

        if name is not None:
            buf = pm.loadBytes(path)
            self.expressions[name] = L2DExpressionMotion.loadJson(buf)

    def loadPose(self, path) -> L2DPose:
        pm = Live2DFramework.getPlatformManager()
        if self.debugMode:
            pm.log("Load Pose : " + path)
        buf = pm.loadBytes(path)
        self.pose = L2DPose.load(buf)
        return self.pose

    def loadPhysics(self, path):
        pm = Live2DFramework.getPlatformManager()
        if self.debugMode:
            pm.log("Load Physics : " + path)
        buf = pm.loadBytes(path)
        self.physics = L2DPhysics.load(buf)

    def hitTestSimple(self, drawID, testX, testY):
        draw_index = self.live2DModel.getDrawDataIndex(drawID)
        if draw_index < 0:
            return False
        points = self.live2DModel.getTransformedPoints(draw_index)
        left = self.live2DModel.getCanvasWidth()
        right = 0
        top = self.live2DModel.getCanvasHeight()
        bottom = 0
        for j in range(0, len(points), 2):
            x = points[j]
            y = points[j + 1]
            if x < left:
                left = x
            if x > right:
                right = x
            if y < top:
                top = y
            if y > bottom:
                bottom = y

        tx = self.modelMatrix.invertTransformX(testX)
        ty = self.modelMatrix.invertTransformY(testY)
        return left <= tx <= right and top <= ty <= bottom



import math
from typing import TYPE_CHECKING, List, Optional

from .DEF import PIVOT_TABLE_SIZE, MAX_INTERPOLATION
from .draw import IDrawData
from .graphics import ClippingManagerOpenGL
from .id import Id
from .type import Array, Float32Array, Int16Array

if TYPE_CHECKING:
    from .draw import MeshContext, Mesh
    from .model import PartsDataContext
    from .graphics import DrawParamOpenGL


class ModelContext:
    NOT_USED_ORDER = -1
    NO_NEXT = -1
    DEFAULT_PARAM_UPDATE_FLAG = False
    PARAM_UPDATED = True
    PARAM_FLOAT_MIN = (-1000000)
    PARAM_FLOAT_MAX = 1000000
    DEFAULT_ARRAY_LENGTH = 0

    def __init__(self, model):
        self.needSetup = True
        self.initVersion = -1
        self.nextParamPos = 0
        self.paramIdList = Array(ModelContext.DEFAULT_ARRAY_LENGTH)
        self.paramValues = Float32Array(ModelContext.DEFAULT_ARRAY_LENGTH)
        self.lastParamValues = Float32Array(ModelContext.DEFAULT_ARRAY_LENGTH)
        self.paramMinValues = Float32Array(ModelContext.DEFAULT_ARRAY_LENGTH)
        self.paramMaxValues = Float32Array(ModelContext.DEFAULT_ARRAY_LENGTH)
        self.savedParamValues = Float32Array(ModelContext.DEFAULT_ARRAY_LENGTH)
        self.updatedParamFlags = Array(ModelContext.DEFAULT_ARRAY_LENGTH)
        self.deformerList = Array()
        self.drawDataList: List[Optional['Mesh']] = Array()
        self.tmpDrawDataList = None
        self.partsDataList = Array()
        self.deformerContextList = Array()
        self.drawContextList = Array()
        self.partsContextList: List[Optional[PartsDataContext]] = Array()
        self.orderList_firstDrawIndex = None
        self.orderList_lastDrawIndex = None
        self.nextList_drawIndex = None
        self.tmpPivotTableIndices = Int16Array(PIVOT_TABLE_SIZE)
        self.tempTArray = Float32Array(MAX_INTERPOLATION)
        self.model = model
        self.clipManager = None
        self.dpGL = None

    def getDrawDataIndex(self, drawDataId) -> int:
        for aH in range(len(self.drawDataList) - 1, 0 - 1, -1):
            if self.drawDataList[aH] is not None and self.drawDataList[aH].getId() == drawDataId:
                return aH

        return -1

    def getDrawData(self, aH):
        if isinstance(aH, Id):
            if self.tmpDrawDataList is None:
                self.tmpDrawDataList = {}
                count = len(self.drawDataList)
                for i in range(0, count, 1):
                    dd = self.drawDataList[i]
                    dd_id = dd.getId()
                    if dd_id is None:
                        continue

                    self.tmpDrawDataList[dd_id] = dd

            return self.tmpDrawDataList[id]
        else:
            if aH < len(self.drawDataList):
                return self.drawDataList[aH]
            else:
                return None

    def release(self):
        self.deformerList.clear()
        self.drawDataList.clear()
        self.partsDataList.clear()
        if self.tmpDrawDataList is not None:
            self.tmpDrawDataList.clear()

        self.deformerContextList.clear()
        self.drawContextList.clear()
        self.partsContextList.clear()

    def init(self):
        self.initVersion += 1
        if len(self.partsDataList) > 0:
            self.release()

        aO = self.model.getModelImpl()
        parts_data_list = aO.getPartsDataList()
        aS = len(parts_data_list)
        aH = Array()
        a3 = Array()
        for aV in range(0, aS, 1):
            a4 = parts_data_list[aV]
            self.partsDataList.append(a4)
            self.partsContextList.append(a4.init())
            base_data_list = a4.getDeformer()
            aR = len(base_data_list)
            for aU in range(0, aR, 1):
                aH.append(base_data_list[aU])

            for aU in range(0, aR, 1):
                aM = base_data_list[aU].init(self)
                aM.setPartsIndex(aV)
                a3.append(aM)

            a1 = a4.getDrawData()
            aP = len(a1)
            for aU in range(0, aP, 1):
                aZ = a1[aU]
                a0 = aZ.init(self)
                a0.partsIndex = aV
                self.drawDataList.append(aZ)
                self.drawContextList.append(a0)

        aY = len(aH)
        aN = Id.DST_BASE_ID()
        while True:
            aX = False
            for aV in range(0, aY, 1):
                aL = aH[aV]
                if aL is None:
                    continue

                a2 = aL.getTargetId()
                if a2 is None or a2 == aN or self.getDeformerIndex(a2) >= 0:
                    self.deformerList.append(aL)
                    self.deformerContextList.append(a3[aV])
                    aH[aV] = None
                    aX = True

            if not aX:
                break

        aI = aO.getParamDefSet()
        if aI is not None:
            aJ = aI.getParamDefFloatList()
            if aJ is not None:
                aW = len(aJ)
                for aV in range(0, aW, 1):
                    aQ = aJ[aV]
                    if aQ is None:
                        continue

                    self.extendAndAddParam(aQ.getParamID(), aQ.getDefaultValue(), aQ.getMinValue(), aQ.getMaxValue())

        self.clipManager = ClippingManagerOpenGL(self.dpGL)
        self.clipManager.init(self, self.drawDataList, self.drawContextList)
        self.needSetup = True

    def update(self):
        aK = len(self.paramValues)
        for i in range(0, aK, 1):
            if self.paramValues[i] != self.lastParamValues[i]:
                self.updatedParamFlags[i] = ModelContext.PARAM_UPDATED
                self.lastParamValues[i] = self.paramValues[i]

        aX = False
        aQ = len(self.deformerList)
        aN = len(self.drawDataList)
        aS = IDrawData.getTotalMinOrder()
        aZ = IDrawData.getTotalMaxOrder()
        aU = aZ - aS + 1
        if self.orderList_firstDrawIndex is None or len(self.orderList_firstDrawIndex) < aU:
            self.orderList_firstDrawIndex = Int16Array(aU)
            self.orderList_lastDrawIndex = Int16Array(aU)

        for i in range(0, aU, 1):
            self.orderList_firstDrawIndex[i] = ModelContext.NOT_USED_ORDER
            self.orderList_lastDrawIndex[i] = ModelContext.NOT_USED_ORDER

        if self.nextList_drawIndex is None or len(self.nextList_drawIndex) < aN:
            self.nextList_drawIndex = Int16Array(aN)

        for i in range(0, aN, 1):
            self.nextList_drawIndex[i] = ModelContext.NO_NEXT

        aL = None
        for aV in range(0, aQ, 1):
            aJ = self.deformerList[aV]
            aH = self.deformerContextList[aV]
            aJ.setupInterpolate(self, aH)
            aJ.setupTransform(self, aH)

        if aL is not None:
            raise RuntimeError

        aR = None
        for aO in range(0, aN, 1):
            aM = self.drawDataList[aO]
            aI = self.drawContextList[aO]
            aM.setupInterpolate(self, aI)
            if aI.isParamOutside():
                continue

            aM.setupTransform(self, aI)
            aT = math.floor(aM.getDrawOrder(aI) - aS)

            aP = self.orderList_lastDrawIndex[aT]

            if aP == ModelContext.NOT_USED_ORDER:
                self.orderList_firstDrawIndex[aT] = aO
            else:
                self.nextList_drawIndex[aP] = aO

            self.orderList_lastDrawIndex[aT] = aO

        if aR is not None:
            raise RuntimeError

        for i in range(len(self.updatedParamFlags) - 1, -1, -1):
            self.updatedParamFlags[i] = ModelContext.DEFAULT_PARAM_UPDATE_FLAG

        self.needSetup = False
        return aX

    def preDraw(self, aH: 'DrawParamOpenGL'):
        if self.clipManager is not None:
            aH.setupDraw()
            self.clipManager.setupClip(self, aH)

    def draw(self, aM):
        if self.orderList_firstDrawIndex is None:
            print("call Ri_.update() before Ri_.draw() ")
            return

        aP = len(self.orderList_firstDrawIndex)
        aM.setupDraw()
        for aK in range(0, aP, 1):
            aN = self.orderList_firstDrawIndex[aK]
            if aN == ModelContext.NOT_USED_ORDER:
                continue

            while True:
                aH = self.drawDataList[aN]
                aI = self.drawContextList[aN]
                if aI.isAvailable():
                    aJ = aI.partsIndex
                    aL = self.partsContextList[aJ]
                    aI.partsOpacity = aL.getPartsOpacity()
                    aH.draw(aM, self, aI)

                aO = self.nextList_drawIndex[aN]
                if aO <= aN or aO == ModelContext.NO_NEXT:
                    break

                aN = aO

    def getParamIndex(self, paramId):
        for i in range(0, len(self.paramIdList), 1):
            p = self.paramIdList[i]
            if p == paramId:
                return i

        return self.extendAndAddParam(paramId, 0, ModelContext.PARAM_FLOAT_MIN, ModelContext.PARAM_FLOAT_MAX)

    def getDeformerIndex(self, aH):
        for aI in range(len(self.deformerList) - 1, 0 - 1, -1):
            if self.deformerList[aI] is not None and self.deformerList[aI].getId() == aH:
                return aI

        return -1

    def extendAndAddParam(self, param_id, default_val, max_val, min_val):
        self.paramIdList.append(param_id)
        self.paramValues.append(default_val)
        self.lastParamValues.append(default_val)
        self.paramMinValues.append(max_val)
        self.paramMaxValues.append(min_val)
        self.updatedParamFlags.append(ModelContext.DEFAULT_PARAM_UPDATE_FLAG)
        ret = self.nextParamPos
        self.nextParamPos += 1
        return ret

    def setDeformer(self, aI, aH):
        self.deformerList[aI] = aH

    def setParamFloat(self, aH, aI):
        if aI < self.paramMinValues[aH]:
            aI = self.paramMinValues[aH]

        if aI > self.paramMaxValues[aH]:
            aI = self.paramMaxValues[aH]

        self.paramValues[aH] = aI

    def loadParam(self):
        for idx, i in enumerate(self.savedParamValues):
            self.paramValues[idx] = i

    def saveParam(self):
        size = len(self.savedParamValues)
        for idx, i in enumerate(self.paramValues):
            if idx >= size:
                self.savedParamValues.append(i)
                continue

            self.savedParamValues[idx] = i

    def getInitVersion(self):
        return self.initVersion

    def requireSetup(self) -> bool:
        return self.needSetup

    def isParamUpdated(self, index) -> bool:
        return self.updatedParamFlags[index] == ModelContext.PARAM_UPDATED

    def getTempPivotTableIndices(self):
        return self.tmpPivotTableIndices

    def getTempT(self):
        return self.tempTArray

    def getDeformer(self, aH):
        return self.deformerList[aH]

    def getParamFloat(self, aH):
        return self.paramValues[aH]

    def getParamMax(self, aH):
        return self.paramMaxValues[aH]

    def getParamMin(self, aH):
        return self.paramMinValues[aH]

    def setPartsOpacity(self, aJ, aH):
        aI = self.partsContextList[aJ]
        aI.setPartsOpacity(aH)

    def getPartsOpacity(self, aI):
        aH = self.partsContextList[aI]
        return aH.getPartsOpacity()

    def getPartsDataIndex(self, aI):
        for aH in range(len(self.partsDataList) - 1, 0 - 1, -1):
            if self.partsDataList[aH] is not None and self.partsDataList[aH].getId() == aI:
                return aH

        return -1

    def getDeformerContext(self, aH):
        return self.deformerContextList[aH]

    def getDrawContext(self, aH) -> 'MeshContext':
        return self.drawContextList[aH]

    def getPartsContext(self, aH) -> 'PartsDataContext':
        return self.partsContextList[aH]

    def setPartMultiplyColor(self, aH, r, g, b, a):
        aI = self.partsContextList[aH]
        aI.multiplyColor[0] = r
        aI.multiplyColor[1] = g
        aI.multiplyColor[2] = b
        aI.multiplyColor[3] = a

    def getPartMultiplyColor(self, aH):
        return self.partsContextList[aH].multiplyColor

    def setPartScreenColor(self, aH, r, g, b, a):
        aI = self.partsContextList[aH]
        aI.screenColor[0] = r
        aI.screenColor[1] = g
        aI.screenColor[2] = b
        aI.screenColor[3] = a

    def getPartScreenColor(self, aH):
        return self.partsContextList[aH].screenColor

    def setDrawParam(self, aH):
        self.dpGL = aH

    def getDrawParam(self):
        return self.dpGL



from .core import Live2D, Live2DGLWrapper
from .core import log as __log
from .framework import Live2DFramework
from .lapp_define import MotionGroup, MotionPriority, HitArea
from .lapp_model import LAppModel
from .params import Parameter, StandardParams
from .platform_manager import PlatformManager


def init():
    Live2D.init()
    Live2DFramework.setPlatformManager(PlatformManager())


def clearBuffer(r=0.0, g=0.0, b=0.0, a=0.0):
    Live2DGLWrapper.clearColor(r, g, b, a)
    Live2DGLWrapper.clear(Live2DGLWrapper.COLOR_BUFFER_BIT)


def setLogEnable(enable: bool):
    __log.setLogEnable(enable)


def logEnable() -> bool:
    return __log.logEnable()


def glewInit():
    pass


def dispose():
    pass


LIVE2D_VERSION = 2

__all__ = ['LAppModel',
           'MotionPriority',
           'MotionGroup',
           "HitArea",
           "StandardParams",
           "init",
           "glewInit",
           "clearBuffer",
           "dispose"]



from typing import Any
from .params import Parameter


def init() -> None:
    """
    initialize inner memory allocator for live2d models
    """
    ...


def dispose() -> None:
    """
    dispose Cubism Framework when no longer using live2d
    """
    ...


def glewInit() -> None:
    """
    initialize inner opengl functions
    """
    ...

def glInit() -> None:
    """
    initialize inner opengl functions
    """
    ...

def clearBuffer(r=0.0, g=0.0, b=0.0, a=0.0) -> None:
    """
    glClearColor(0.0, 0.0, 0.0, 0.0)

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
    
    glClearDepth(1.0)
    """
    ...


def setLogEnable(enable: bool):
    ...


def logEnable() -> bool:
    ...


class LAppModel:
    """
    The LAppModel class provides a structured way to interact with Live2D models, 
    enabling you to load assets, update the model per frame, manage motions, set 
    expressions, and perform hit testing. 
    """

    def __init__(self):
        ...

    def LoadModelJson(self, fileName: str | Any) -> None:
        """
        Load Live2D model assets.
        
        :param fileName: Name of the model's JSON configuration file.
        """
        ...

    def Resize(self, ww: int | Any, wh: int | Any) -> None:
        """
        adjust model canvas to window size
        :param ww: window width
        :param wh: window height
        :return:
        """
        ...

    def Draw(self) -> None:
        """
        update model shapes with the params set by `LAppModel.Update` and  `LAppModel.SetParameterValue`, and then render them 
        """
        ...

    def StartMotion(self, group: str | Any, no: int | Any, priority: int | Any, onStartMotionHandler=None,
                    onFinishMotionHandler=None) -> None:
        """
        Start a specific motion for the model.
        
        :param group: The group name of the motion.
        :param no: The motion number within the group.
        :param priority: Priority of the motion. Higher priority motions can interrupt lower priority ones.
        :param onStartMotionHandler: Optional callback function that gets called when the motion starts.
        :param onFinishMotionHandler: Optional callback function that gets called when the motion finishes.
        """
        ...

    def StartRandomMotion(self, group: str | Any = None, priority: int | Any = 3, onStartMotionHandler=None,
                          onFinishMotionHandler=None) -> None:
        """
        Start a random motion from a specified group.
        
        :param group: The group name of the motion.
        :param priority: Priority of the motion. Higher priority motions can interrupt lower priority ones.
        :param onFinishedMotionHandler: Optional callback function that gets called when the motion finishes.
        """
        ...

    def SetExpression(self, expressionID: str | Any, fadeout=-1) -> None:
        """
        Set a specific expression for the model.
        
        :param expressionID: name of the expression to be set.
        """
        ...

    def SetRandomExpression(self, fadeout=-1) -> str:
        """
        Set a random expression for the model.
        """
        ...

    def HitTest(self, hitAreaName: str, x: float | Any, y: float | Any) -> bool:
        """
        to test if the current clicked area is the target `HitArea` defined in xxx.model3.json

        x, y are relative to the window topleft
        """
        ...

    def HasMocConsistencyFromFile(self, mocFileName: str | Any) -> bool:
        """
        Check if the model's MOC file is consistent.
        
        :param mocFileName: Name of the MOC file to check.
        :return: True if the MOC file is consistent, otherwise False.
        """
        ...

    def Drag(self, x: float | Any, y: float | Any) -> None:
        """
        :param x: global_mouse_x - window_x
        :param y: global_mouse_y - window_y
        """
        ...

    def IsMotionFinished(self) -> bool:
        """
        当前正在播放的动作是否已经结束
        :return:
        """
        ...

    def SetOffset(self, dx: float | Any, dy: float | Any) -> None:
        """
        设置模型中心坐标的偏移量
        :param dx:
        :param dy:
        :return:
        """
        ...

    def SetScale(self, scale: float | Any) -> None:
        """
        设置模型缩放比例
        :param scale: 缩放比例
        :return:
        """
        ...

    def Rotate(self, degrees: float) -> None:
        """
        counterclockwise when degrees > 0
        """

    def SetParameterValue(self, paramId: str, value: float, weight: float = 1.0) -> None:
        """
        为对应控制参数设置具体值  
        调用时机: 在 CalcParameters 后，在 Update 之前
        :param paramId: live2d 模型内嵌的参数，详细见 live2d.v3.params.StandardParams
        :param value: 所有可操作参数见官方文档: https://docs.live2d.com/en/cubism-editor-manual/standard-parameter-list/
        :param weight: 当前传入的值和原值的比例，最终值=原值*(1-weight)+传入值*weight
        """
        ...

    def SetIndexParamValue(self, index: int, value: float, weight: float = 1.0) -> None:
        ...

    def AddParameterValue(self, paramId: str, value: float) -> None:
        """
        最终值=原值+value
        """
        ...

    def AddIndexParamValue(self, index: int, value: float) -> None:
        ...

    def Update(self) -> None:
        """
        初始化呼吸、动作、姿势、表情、各部分透明度等必要的参数值
        """
        ...

    def SetAutoBreathEnable(self, enable: bool) -> None:
        """
        开启自动呼吸
        """
        ...

    def SetAutoBlinkEnable(self, enable: bool) -> None:
        """
        开启自动眨眼
        """
        ...

    def GetParameterCount(self) -> int:
        ...

    def GetParameter(self, index: int) -> Parameter:
        ...
    
    def GetParamIds(self) -> list[str]:
        ...
    
    def GetParameterValue(self, index: int) -> float:
        ...

    def GetPartCount(self) -> int:
        ...

    def GetPartId(self, index: int) -> str:
        ...

    def GetPartIds(self) -> list[str]:
        ...

    def SetPartOpacity(self, index: int, opacity: float) -> None:
        ...

    def HitPart(self, x: float, y: float, topOnly: bool = False) -> list[str]:
        """
        
        :param x: 屏幕坐标 x
        :param y: 屏幕坐标 y
        :param topOnly: 只返回最顶部的 part id
        :return: part id 列表
        """
        ...

    def SetPartScreenColor(self, partIndex: int, r: float, g: float, b: float, a: float):
        ...
    
    def GetPartScreenColor(self, partIndex: int) -> tuple[float]:
        ...

    def SetPartMultiplyColor(self, partIndex: int, r: float, g: float, b: float, a: float):
        ...
    
    def GetPartMultiplyColor(self, partIndex: int) -> tuple[float]:
        ...

    def ResetExpression(self) -> None:
        """
        重置为默认表情
        """
        ...

    def StopAllMotions(self) -> None:
        ...

    def ResetParameters(self) -> None:
        ...

    def ResetPose(self) -> None:
        ...

    def GetExpressionIds(self) -> list[str]:
        ...

    def GetMotionGroups(self) -> dict[str, int]:
        ...


import time

RESET = "\033[0m"
RED = "\033[31m"
GREEN = "\033[32m"
YELLOW = "\033[33m"
BLUE = "\033[34m"
MAGENTA = "\033[35m"
CYAN = "\033[36m"
WHITE = "\033[37m"


def Debug(*args, **kwargs):
    print(
        time.strftime(f"{BLUE}[DEBUG]"),
        *args,
        RESET,
        **kwargs
    )


def Info(*args, **kwargs):
    print(
        time.strftime("[INFO] "),
        *args,
        **kwargs
    )


def Error(*args, **kwargs):
    print(
        time.strftime(f"{RED}[ERROR]"),
        *args,
        RESET,
        **kwargs
    )



import math
from typing import TYPE_CHECKING, Optional, List

from .deformer import Deformer
from .rotation_context import RotationContext
from ..DEF import LIVE2D_FORMAT_VERSION_V2_10_SDK2
from ..param import PivotManager
from ..type import Float32Array, Array
from ..util import UtMath

if TYPE_CHECKING:
    from .deformer_context import DeformerContext
    from ..model_context import ModelContext
    from ..io import BinaryReader


class RotationDeformer(Deformer):
    # 代替c语言中的指针/引用
    temp1 = [0.0, 0.0]
    temp2 = [0.0, 0.0]
    temp3 = [0.0, 0.0]
    temp4 = [0.0, 0.0]
    temp5 = [0.0, 0.0]
    temp6 = [0.0, 0.0]
    paramOutside = [False]

    def __init__(self):
        super().__init__()
        self.pivotManager: Optional['PivotManager'] = None
        self.affines: Optional[List['AffineEnt']] = None

    def getType(self) -> int:
        return Deformer.TYPE_ROTATION

    def read(self, br: 'BinaryReader'):
        super().read(br)
        self.pivotManager = br.readObject()
        self.affines = br.readObject()
        super().readOpacity(br)

    def init(self, mc) -> 'RotationContext':
        rctx = RotationContext(self)
        rctx.interpolatedAffine = AffineEnt()
        if self.needTransform():
            rctx.transformedAffine = AffineEnt()

        return rctx

    def setupInterpolate(self, mctx: 'ModelContext', rctx: 'RotationContext'):
        if not (self == rctx.getDeformer()):
            raise RuntimeError("context not match")

        if not self.pivotManager.checkParamUpdated(mctx):
            return

        success = RotationDeformer.paramOutside
        success[0] = False
        a2 = self.pivotManager.calcPivotValues(mctx, success)
        rctx.setOutsideParam(success[0])
        self.interpolateOpacity(mctx, self.pivotManager, rctx, success)
        a3 = mctx.getTempPivotTableIndices()
        ba = mctx.getTempT()
        self.pivotManager.calcPivotIndices(a3, ba, a2)
        if a2 <= 0:
            bn_3 = self.affines[a3[0]]
            rctx.interpolatedAffine.init(bn_3)
        else:
            if a2 == 1:
                bn_1 = self.affines[a3[0]]
                bl = self.affines[a3[1]]
                a9 = ba[0]
                rctx.interpolatedAffine.originX = bn_1.originX + (bl.originX - bn_1.originX) * a9
                rctx.interpolatedAffine.originY = bn_1.originY + (bl.originY - bn_1.originY) * a9
                rctx.interpolatedAffine.scaleX = bn_1.scaleX + (bl.scaleX - bn_1.scaleX) * a9
                rctx.interpolatedAffine.scaleY = bn_1.scaleY + (bl.scaleY - bn_1.scaleY) * a9
                rctx.interpolatedAffine.rotationDeg = bn_1.rotationDeg + (
                        bl.rotationDeg - bn_1.rotationDeg) * a9
            else:
                if a2 == 2:
                    bn_1 = self.affines[a3[0]]
                    bl = self.affines[a3[1]]
                    a1 = self.affines[a3[2]]
                    a0 = self.affines[a3[3]]
                    a9 = ba[0]
                    a8 = ba[1]
                    bC = bn_1.originX + (bl.originX - bn_1.originX) * a9
                    bB = a1.originX + (a0.originX - a1.originX) * a9
                    rctx.interpolatedAffine.originX = bC + (bB - bC) * a8
                    bC = bn_1.originY + (bl.originY - bn_1.originY) * a9
                    bB = a1.originY + (a0.originY - a1.originY) * a9
                    rctx.interpolatedAffine.originY = bC + (bB - bC) * a8
                    bC = bn_1.scaleX + (bl.scaleX - bn_1.scaleX) * a9
                    bB = a1.scaleX + (a0.scaleX - a1.scaleX) * a9
                    rctx.interpolatedAffine.scaleX = bC + (bB - bC) * a8
                    bC = bn_1.scaleY + (bl.scaleY - bn_1.scaleY) * a9
                    bB = a1.scaleY + (a0.scaleY - a1.scaleY) * a9
                    rctx.interpolatedAffine.scaleY = bC + (bB - bC) * a8
                    bC = bn_1.rotationDeg + (bl.rotationDeg - bn_1.rotationDeg) * a9
                    bB = a1.rotationDeg + (a0.rotationDeg - a1.rotationDeg) * a9
                    rctx.interpolatedAffine.rotationDeg = bC + (bB - bC) * a8
                else:
                    if a2 == 3:
                        aP = self.affines[a3[0]]
                        aO = self.affines[a3[1]]
                        bu = self.affines[a3[2]]
                        bs = self.affines[a3[3]]
                        aK = self.affines[a3[4]]
                        aJ = self.affines[a3[5]]
                        bj = self.affines[a3[6]]
                        bi = self.affines[a3[7]]
                        a9 = ba[0]
                        a8 = ba[1]
                        a6 = ba[2]
                        bC = aP.originX + (aO.originX - aP.originX) * a9
                        bB = bu.originX + (bs.originX - bu.originX) * a9
                        bz = aK.originX + (aJ.originX - aK.originX) * a9
                        by = bj.originX + (bi.originX - bj.originX) * a9
                        rctx.interpolatedAffine.originX = (1 - a6) * (bC + (bB - bC) * a8) + a6 * (
                                bz + (by - bz) * a8)
                        bC = aP.originY + (aO.originY - aP.originY) * a9
                        bB = bu.originY + (bs.originY - bu.originY) * a9
                        bz = aK.originY + (aJ.originY - aK.originY) * a9
                        by = bj.originY + (bi.originY - bj.originY) * a9
                        rctx.interpolatedAffine.originY = (1 - a6) * (bC + (bB - bC) * a8) + a6 * (
                                bz + (by - bz) * a8)
                        bC = aP.scaleX + (aO.scaleX - aP.scaleX) * a9
                        bB = bu.scaleX + (bs.scaleX - bu.scaleX) * a9
                        bz = aK.scaleX + (aJ.scaleX - aK.scaleX) * a9
                        by = bj.scaleX + (bi.scaleX - bj.scaleX) * a9
                        rctx.interpolatedAffine.scaleX = (1 - a6) * (bC + (bB - bC) * a8) + a6 * (
                                bz + (by - bz) * a8)
                        bC = aP.scaleY + (aO.scaleY - aP.scaleY) * a9
                        bB = bu.scaleY + (bs.scaleY - bu.scaleY) * a9
                        bz = aK.scaleY + (aJ.scaleY - aK.scaleY) * a9
                        by = bj.scaleY + (bi.scaleY - bj.scaleY) * a9
                        rctx.interpolatedAffine.scaleY = (1 - a6) * (bC + (bB - bC) * a8) + a6 * (
                                bz + (by - bz) * a8)
                        bC = aP.rotationDeg + (aO.rotationDeg - aP.rotationDeg) * a9
                        bB = bu.rotationDeg + (bs.rotationDeg - bu.rotationDeg) * a9
                        bz = aK.rotationDeg + (aJ.rotationDeg - aK.rotationDeg) * a9
                        by = bj.rotationDeg + (bi.rotationDeg - bj.rotationDeg) * a9
                        rctx.interpolatedAffine.rotationDeg = (1 - a6) * (bC + (bB - bC) * a8) + a6 * (
                                bz + (by - bz) * a8)
                    else:
                        if a2 == 4:
                            aT = self.affines[a3[0]]
                            aS = self.affines[a3[1]]
                            bE = self.affines[a3[2]]
                            bD = self.affines[a3[3]]
                            aN = self.affines[a3[4]]
                            aM = self.affines[a3[5]]
                            bp = self.affines[a3[6]]
                            bo = self.affines[a3[7]]
                            bh = self.affines[a3[8]]
                            bg = self.affines[a3[9]]
                            aY = self.affines[a3[10]]
                            aW = self.affines[a3[11]]
                            a7 = self.affines[a3[12]]
                            a5 = self.affines[a3[13]]
                            aR = self.affines[a3[14]]
                            aQ = self.affines[a3[15]]
                            a9 = ba[0]
                            a8 = ba[1]
                            a6 = ba[2]
                            a4 = ba[3]
                            bC = aT.originX + (aS.originX - aT.originX) * a9
                            bB = bE.originX + (bD.originX - bE.originX) * a9
                            bz = aN.originX + (aM.originX - aN.originX) * a9
                            by = bp.originX + (bo.originX - bp.originX) * a9
                            bv = bh.originX + (bg.originX - bh.originX) * a9
                            bt = aY.originX + (aW.originX - aY.originX) * a9
                            br = a7.originX + (a5.originX - a7.originX) * a9
                            bq = aR.originX + (aQ.originX - aR.originX) * a9
                            rctx.interpolatedAffine.originX = (1 - a4) * (
                                    (1 - a6) * (bC + (bB - bC) * a8) + a6 * (bz + (by - bz) * a8)) + a4 * (
                                                                      (1 - a6) * (
                                                                      bv + (bt - bv) * a8) + a6 * (
                                                                              br + (bq - br) * a8))
                            bC = aT.originY + (aS.originY - aT.originY) * a9
                            bB = bE.originY + (bD.originY - bE.originY) * a9
                            bz = aN.originY + (aM.originY - aN.originY) * a9
                            by = bp.originY + (bo.originY - bp.originY) * a9
                            bv = bh.originY + (bg.originY - bh.originY) * a9
                            bt = aY.originY + (aW.originY - aY.originY) * a9
                            br = a7.originY + (a5.originY - a7.originY) * a9
                            bq = aR.originY + (aQ.originY - aR.originY) * a9
                            rctx.interpolatedAffine.originY = (1 - a4) * (
                                    (1 - a6) * (bC + (bB - bC) * a8) + a6 * (bz + (by - bz) * a8)) + a4 * (
                                                                      (1 - a6) * (
                                                                      bv + (bt - bv) * a8) + a6 * (
                                                                              br + (bq - br) * a8))
                            bC = aT.scaleX + (aS.scaleX - aT.scaleX) * a9
                            bB = bE.scaleX + (bD.scaleX - bE.scaleX) * a9
                            bz = aN.scaleX + (aM.scaleX - aN.scaleX) * a9
                            by = bp.scaleX + (bo.scaleX - bp.scaleX) * a9
                            bv = bh.scaleX + (bg.scaleX - bh.scaleX) * a9
                            bt = aY.scaleX + (aW.scaleX - aY.scaleX) * a9
                            br = a7.scaleX + (a5.scaleX - a7.scaleX) * a9
                            bq = aR.scaleX + (aQ.scaleX - aR.scaleX) * a9
                            rctx.interpolatedAffine.scaleX = (1 - a4) * (
                                    (1 - a6) * (bC + (bB - bC) * a8) + a6 * (bz + (by - bz) * a8)) + a4 * (
                                                                     (1 - a6) * (
                                                                     bv + (bt - bv) * a8) + a6 * (
                                                                             br + (bq - br) * a8))
                            bC = aT.scaleY + (aS.scaleY - aT.scaleY) * a9
                            bB = bE.scaleY + (bD.scaleY - bE.scaleY) * a9
                            bz = aN.scaleY + (aM.scaleY - aN.scaleY) * a9
                            by = bp.scaleY + (bo.scaleY - bp.scaleY) * a9
                            bv = bh.scaleY + (bg.scaleY - bh.scaleY) * a9
                            bt = aY.scaleY + (aW.scaleY - aY.scaleY) * a9
                            br = a7.scaleY + (a5.scaleY - a7.scaleY) * a9
                            bq = aR.scaleY + (aQ.scaleY - aR.scaleY) * a9
                            rctx.interpolatedAffine.scaleY = (1 - a4) * (
                                    (1 - a6) * (bC + (bB - bC) * a8) + a6 * (bz + (by - bz) * a8)) + a4 * (
                                                                     (1 - a6) * (
                                                                     bv + (bt - bv) * a8) + a6 * (
                                                                             br + (bq - br) * a8))
                            bC = aT.rotationDeg + (aS.rotationDeg - aT.rotationDeg) * a9
                            bB = bE.rotationDeg + (bD.rotationDeg - bE.rotationDeg) * a9
                            bz = aN.rotationDeg + (aM.rotationDeg - aN.rotationDeg) * a9
                            by = bp.rotationDeg + (bo.rotationDeg - bp.rotationDeg) * a9
                            bv = bh.rotationDeg + (bg.rotationDeg - bh.rotationDeg) * a9
                            bt = aY.rotationDeg + (aW.rotationDeg - aY.rotationDeg) * a9
                            br = a7.rotationDeg + (a5.rotationDeg - a7.rotationDeg) * a9
                            bq = aR.rotationDeg + (aQ.rotationDeg - aR.rotationDeg) * a9
                            rctx.interpolatedAffine.rotationDeg = (1 - a4) * (
                                    (1 - a6) * (bC + (bB - bC) * a8) + a6 * (bz + (by - bz) * a8)) + a4 * (
                                                                          (1 - a6) * (
                                                                          bv + (bt - bv) * a8) + a6 * (
                                                                                  br + (bq - br) * a8))
                        else:
                            aV = int(pow(2, a2))
                            aZ = Float32Array(aV)
                            for bk in range(0, aV, 1):
                                aI = bk
                                aH = 1
                                for aL in range(0, a2, 1):
                                    aH *= (1 - ba[aL]) if (aI % 2 == 0) else ba[aL]
                                    aI /= 2

                                aZ[bk] = aH

                            bA = Array()
                            for aU in range(0, aV, 1):
                                bA[aU] = self.affines[a3[aU]]

                            be = 0
                            bc = 0
                            bd = 0
                            bb = 0
                            aX = 0
                            for aU in range(0, aV, 1):
                                be += aZ[aU] * bA[aU].originX
                                bc += aZ[aU] * bA[aU].originY
                                bd += aZ[aU] * bA[aU].scaleX
                                bb += aZ[aU] * bA[aU].scaleY
                                aX += aZ[aU] * bA[aU].rotationDeg

                            rctx.interpolatedAffine.originX = be
                            rctx.interpolatedAffine.originY = bc
                            rctx.interpolatedAffine.scaleX = bd
                            rctx.interpolatedAffine.scaleY = bb
                            rctx.interpolatedAffine.rotationDeg = aX

        bn = self.affines[a3[0]]
        rctx.interpolatedAffine.reflectX = bn.reflectX
        rctx.interpolatedAffine.reflectY = bn.reflectY

    def setupTransform(self, mctx: 'ModelContext', rctx: 'RotationContext'):
        if not (self == rctx.getDeformer()):
            raise RuntimeError("Invalid Deformer")

        rctx.setAvailable(True)
        if not self.needTransform():
            rctx.setTotalScale_notForClient(rctx.interpolatedAffine.scaleX)
            rctx.setTotalOpacity(rctx.getInterpolatedOpacity())
        else:
            aT = self.getTargetId()
            if rctx.tmpDeformerIndex == Deformer.DEFORMER_INDEX_NOT_INIT:
                rctx.tmpDeformerIndex = mctx.getDeformerIndex(aT)

            if rctx.tmpDeformerIndex < 0:
                print("deformer is not reachable")
                rctx.setAvailable(False)
            else:
                deformer = mctx.getDeformer(rctx.tmpDeformerIndex)
                if deformer is not None:
                    dctx = mctx.getDeformerContext(rctx.tmpDeformerIndex)
                    aS = RotationDeformer.temp1
                    aS[0] = rctx.interpolatedAffine.originX
                    aS[1] = rctx.interpolatedAffine.originY
                    aJ = RotationDeformer.temp2
                    aJ[0] = 0
                    aJ[1] = -0.1
                    aO = dctx.getDeformer().getType()
                    if aO == Deformer.TYPE_ROTATION:
                        aJ[1] = -10
                    else:
                        aJ[1] = -0.1

                    aQ = RotationDeformer.temp3
                    self.getDirectionOnDst(mctx, deformer, dctx, aS, aJ, aQ)
                    aP = UtMath.getAngleNotAbs(aJ, aQ)
                    deformer.transformPoints(mctx, dctx, aS, aS, 1, 0, 2)
                    rctx.transformedAffine.originX = aS[0]
                    rctx.transformedAffine.originY = aS[1]
                    rctx.transformedAffine.scaleX = rctx.interpolatedAffine.scaleX
                    rctx.transformedAffine.scaleY = rctx.interpolatedAffine.scaleY
                    rctx.transformedAffine.rotationDeg = rctx.interpolatedAffine.rotationDeg - aP * UtMath.RAD_TO_DEG
                    aK = dctx.getTotalScale()
                    rctx.setTotalScale_notForClient(aK * rctx.transformedAffine.scaleX)
                    aN = dctx.getTotalOpacity()
                    rctx.setTotalOpacity(aN * rctx.getInterpolatedOpacity())
                    rctx.transformedAffine.reflectX = rctx.interpolatedAffine.reflectX
                    rctx.transformedAffine.reflectY = rctx.interpolatedAffine.reflectY
                    rctx.setAvailable(dctx.isAvailable())
                else:
                    rctx.setAvailable(False)

    def transformPoints(self, mc: 'ModelContext', dc: 'RotationContext', srcPoints: List[float], dstPoints: List[float],
                        numPoint: int, ptOffset: int, ptStep: int):
        if not (self == dc.getDeformer()):
            raise RuntimeError("context not match")

        aH = dc
        aU = aH.transformedAffine if aH.transformedAffine is not None else aH.interpolatedAffine
        a0 = math.sin(UtMath.DEG_TO_RAD * aU.rotationDeg)
        aP = math.cos(UtMath.DEG_TO_RAD * aU.rotationDeg)
        a3 = aH.getTotalScale()
        aW = -1 if aU.reflectX else 1
        aV = -1 if aU.reflectY else 1
        aS = aP * a3 * aW
        aQ = -a0 * a3 * aV
        a1 = a0 * a3 * aW
        aZ = aP * a3 * aV
        aY = aU.originX
        aX = aU.originY
        aI = numPoint * ptStep
        for aK in range(ptOffset, aI, ptStep):
            aN = srcPoints[aK]
            aM = srcPoints[aK + 1]
            dstPoints[aK] = aS * aN + aQ * aM + aY
            dstPoints[aK + 1] = a1 * aN + aZ * aM + aX

    @staticmethod
    def getDirectionOnDst(mdc: 'ModelContext', targetToDst: 'Deformer', targetToDstContext: 'DeformerContext', srcOrigin, srcDir, retDir):
        if not (targetToDst == targetToDstContext.getDeformer()):
            raise RuntimeError("context not match")

        aO = RotationDeformer.temp4
        RotationDeformer.temp4[0] = srcOrigin[0]
        RotationDeformer.temp4[1] = srcOrigin[1]
        targetToDst.transformPoints(mdc, targetToDstContext, aO, aO, 1, 0, 2)
        aL = RotationDeformer.temp5
        aS = RotationDeformer.temp6
        aN = 10
        aJ = 1
        for aM in range(0, aN, 1):
            aS[0] = srcOrigin[0] + aJ * srcDir[0]
            aS[1] = srcOrigin[1] + aJ * srcDir[1]
            targetToDst.transformPoints(mdc, targetToDstContext, aS, aL, 1, 0, 2)
            aL[0] -= aO[0]
            aL[1] -= aO[1]
            if aL[0] != 0 or aL[1] != 0:
                retDir[0] = aL[0]
                retDir[1] = aL[1]
                return

            aS[0] = srcOrigin[0] - aJ * srcDir[0]
            aS[1] = srcOrigin[1] - aJ * srcDir[1]
            targetToDst.transformPoints(mdc, targetToDstContext, aS, aL, 1, 0, 2)
            aL[0] -= aO[0]
            aL[1] -= aO[1]
            if aL[0] != 0 or aL[1] != 0:
                aL[0] = -aL[0]
                aL[0] = -aL[0]
                retDir[0] = aL[0]
                retDir[1] = aL[1]
                return

            aJ *= 0.1


        print("Invalid state\n")


class AffineEnt:

    def __init__(self):
        self.originX = 0
        self.originY = 0
        self.scaleX = 1
        self.scaleY = 1
        self.rotationDeg = 0
        self.reflectX = False
        self.reflectY = False

    def init(self, other):
        self.originX = other.originX
        self.originY = other.originY
        self.scaleX = other.scaleX
        self.scaleY = other.scaleY
        self.rotationDeg = other.rotationDeg
        self.reflectX = other.reflectX
        self.reflectY = other.reflectY

    def read(self, br: 'BinaryReader'):
        self.originX = br.readFloat32()
        self.originY = br.readFloat32()
        self.scaleX = br.readFloat32()
        self.scaleY = br.readFloat32()
        self.rotationDeg = br.readFloat32()
        if br.getFormatVersion() >= LIVE2D_FORMAT_VERSION_V2_10_SDK2:
            self.reflectX = br.readBoolean()
            self.reflectY = br.readBoolean()
